'''
Implement XGBoost (a gradient-boosted decision tree method) model for tabular individual outcome
prediction in clinical trials.
'''
import pdb
import os
import joblib
import pickle

from xgboost import XGBClassifier as xgb_clf_model
from xgboost import XGBRegressor as xgb_reg_model
from sklearn.multioutput import MultiOutputClassifier # wrap xgboost classifier to be multilable predictor
import numpy as np
import pandas as pd

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
from .base import TabularIndivBase

class BuildModel:
    def __new__(self, config):
        if config['mode'] in ['binary']:
            config['objective'] = 'binary:logistic'
            model = xgb_clf_model(**config)
        
        elif config['mode'] in ['multiclass']:
            config['objective'] = 'multi:softmax'
            model = xgb_clf_model(**config)

        elif config['mode'] in ['multilabel']:
            config['objective'] = 'binary:logistic'
            xgb_est = xgb_clf_model(**config)
            model = MultiOutputClassifier(xgb_est)

        else:
            config['objective'] = 'reg:squarederror'
            model = xgb_reg_model(**config)

        return model

class XGBoost(TabularIndivBase):
    '''
    Implement XGBoost model for tabular individual outcome
    prediction in clinical trials.

    Parameters
    ----------
    n_estimators : int
        Number of boosting rounds.

    max_depth : int
        Maximum tree depth for base learners.

    mode: str
        The task's objectives, in `binary`, `multiclass`, `multilabel`, or `regression`.
        Do not support early stopping when `multilabel`.

    n_jobs: int
        Number of parallel threads used to run xgboost. When used with other Scikit-Learn algorithms like grid search, 
        you may choose which algorithm to parallelize and balance the threads. Creating thread contention will significantly slow down both algorithms.

    reg_alpha: float
        L1 regularization term on weights (xgb's alpha).
    
    reg_lambda: float
        L2 regularization term on weights (xgb's lambda).

    experiment_id: str, optional (default='test')
        The name of current experiment. Decide the saved model checkpoint name.
    '''
    def __init__(self,
        mode,
        n_estimators=100,
        max_depth=8,
        n_jobs=0,
        reg_alpha=0,
        reg_lambda=0,
        num_class=None,
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id=experiment_id)
        if mode in ['multclass']:
            assert num_class is not None, 'Should specify `num_class` if `mode` is `multiclass`!'

        self.config = {
            'mode':mode,
            'n_estimators': n_estimators,
            'max_depth': max_depth,
            'reg_alpha': reg_alpha,
            'experiment_id': experiment_id,
            'reg_lambda': reg_lambda,
            'model_name': 'xgboost',
            'n_jobs': n_jobs,
            'num_class':num_class,
        }
        self._save_config(self.config)

    def fit(self, train_data, valid_data=None):
        '''Train logistic regression model to predict patient outcome
        with tabular input data.

        Parameters
        ----------
        train_data: dict
            {
            'x': TabularPatientBase or pd.DataFrame,
            'y': pd.Series or np.ndarray
            }

            - 'x' contain all patient features; 
            - 'y' contain labels for each row.

        valid_data: same as train_data.
            Validation set for early stopping.
        '''
        self._input_data_check(train_data)
        x_feat, y = self._parse_input_data(train_data)
        eval_set = None

        if valid_data is not None and self.config['mode'] != 'multilabel':
            self._input_data_check(train_data)
            x_val, y_val = self._parse_input_data(valid_data)
            eval_set = [(x_val, y_val)]

        self._build_model()
        self._fit_model(x_feat, y, eval_set=eval_set)

    def predict(self, test_data):
        '''
        Make prediction probability based on the learned model.
        Save to `self.result_dir`.

        Parameters
        ----------
        test_data: dict
            {
            'x': TabularPatientBase or pd.DataFrame,
            'y': pd.Series or np.ndarray
            }
            
            - 'x' contain all patient features; 
            - 'y' contain labels for each row. Ignored for prediction function.

        Returns
        -------
        ypred: np.ndarray
            For binary classification, return shape (n, );
            For multiclass classification, return shape (n, n_class).

        '''
        self._input_data_check(test_data)
        x_feat, y = self._parse_input_data(test_data)
        ypred = self.model.predict_proba(x_feat)

        if isinstance(ypred, list): # multilabel prediction
            ypred = np.stack(ypred,0)
            ypred = ypred[...,1].T # n_sample, n_labels

        if ypred.shape[1] == 2: # binary
            ypred = ypred[:,1]

        # save results to dir
        pickle.dump(ypred, open(os.path.join(self.result_dir, 'pred.pkl'), 'wb'))
        if isinstance(test_data, dict):
            if 'y' in test_data:
                pickle.dump(test_data['y'], open(os.path.join(self.result_dir, 'label.pkl'), 'wb'))

        return ypred

    def save_model(self, output_dir=None):
        '''
        Save the learned XGBoost model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is None:
            output_dir = self.checkout_dir
        make_dir_if_not_exist(output_dir)

        self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'indiv-tabular.model')
        joblib.dump(self.model, ckpt_path)

    def load_model(self, checkpoint=None):
        '''
        Load the learned XGBoost model from the disk.

        Parameters
        ----------
        checkpoint: str or None
            - If a directory, the only checkpoint file `.model` will be loaded.
            - If a filepath, will load from this file;
            - If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir

        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)
        self.config = self._load_config(config_filename)

    def _build_model(self):
        self.model = BuildModel(self.config)

    def _fit_model(self, x_feat, y, eval_set):
        self.model.fit(x_feat, y, eval_set=eval_set)

    def _parse_input_data(self, inputs):
        if isinstance(inputs, TabularPatientBase):
            x_feat = inputs.df
            y = None

        if isinstance(inputs, pd.DataFrame):
            x_feat = inputs
            y = None

        if isinstance(inputs, dict):
            if isinstance(inputs['x'], TabularPatientBase):
                dataset = inputs['x']
                x_feat = dataset.df
            else:
                x_feat = inputs['x']
            y = inputs['y'] if 'y' in inputs else None

        return x_feat, y


import torch
import pandas as pd
from torch import nn
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence
import numpy as np
from collections import defaultdict
from pytrial.data.patient_data import SequencePatientBase, SeqPatientCollator
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from pytrial.tasks.trial_simulation.sequence.base import SequenceSimulationBase
from pytrial.tasks.trial_simulation.data import SequencePatient
import os
import pdb
import json
import torch.nn.functional as F
import random
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from numpy import vstack
from torch.optim import Adam

class trial_data(Dataset):
    # load the dataset
    def __init__(self, X, y):
        self.X = X
        self.y = y

        self.X = self.X.astype('float32')
        self.y = self.y.astype('float32')
        #self.y = self.y.reshape((len(self.y), 1))

    # number of rows in the dataset
    def __len__(self):
        return len(self.X)

    # get a row at an index
    def __getitem__(self, idx):
        return [self.X.iloc[idx, :].to_numpy(), self.y.iloc[idx, :].to_numpy()]

def prepare_data(X_train, y_train, batch_size =64):
    # load the dataset
    train = trial_data(X_train, y_train)
    #test = trial_data(X_test, y_test)
    
    batch_size =batch_size
    # prepare data loaders. cannot shuffle during training
    train_dl = DataLoader(train, batch_size=batch_size, shuffle=False)#, drop_last=True)
    #test_dl = DataLoader(test, batch_size=batch_size, shuffle=False)
    return train_dl#, test_dl

class Encoder(nn.Module):
    
    def __init__(self, vocab_size, event_order, hidden_dim, latent_dim):
        super(Encoder, self).__init__()
        input_dim = vocab_size[event_order]
        self.FC_input = nn.Linear(input_dim, hidden_dim)
        torch.nn.init.xavier_uniform_(self.FC_input.weight)
        self.FC_mean  = nn.Linear(hidden_dim, latent_dim)
        self.FC_var   = nn.Linear (hidden_dim, latent_dim)
        torch.nn.init.xavier_uniform_(self.FC_mean.weight)
        torch.nn.init.xavier_uniform_(self.FC_var.weight)
        
        self.LeakyReLU = nn.ReLU()
        self.training = True
        
    def forward(self, x):
        h_       = self.LeakyReLU(self.FC_input(x))
        mean     = self.FC_mean(h_)
        log_var  = self.FC_var(h_)
        return mean, log_var

class Decoder(nn.Module):
    def __init__(self, latent_dim, hidden_dim, vocab_size, event_order):
        super(Decoder, self).__init__()
        output_dim = vocab_size[event_order]
        self.FC_hidden = nn.Linear(latent_dim, hidden_dim)
        torch.nn.init.xavier_uniform_(self.FC_hidden.weight)
        self.FC_output = nn.Linear(hidden_dim, output_dim)
        torch.nn.init.xavier_uniform_(self.FC_output.weight)
        
        self.LeakyReLU = nn.ReLU()
        
    def forward(self, x):
        h     = self.LeakyReLU(self.FC_hidden(x))
        x_hat = torch.sigmoid(self.FC_output(h))
        return x_hat

class AE_pred_module(nn.Module):
    def __init__(self, latent_dim, hidden_dim, vocab_size, treatment_order, event_order):
        super(AE_pred_module, self).__init__()
        treatment_dim = vocab_size[treatment_order]
        ae_dim = vocab_size[event_order]
        self.FC_hidden = nn.Linear(latent_dim+treatment_dim, latent_dim+10)
        self.FC_hidden2 = nn.Linear(latent_dim +10, hidden_dim)
        self.FC_hidden3 = nn.Linear(hidden_dim, hidden_dim)
        self.FC_output = nn.Linear(hidden_dim, ae_dim)
        self.FC_output1 = nn.Linear(ae_dim, ae_dim)
        self.LeakyReLU= nn.ReLU()
        self.bn1 = nn.BatchNorm1d(hidden_dim)
    
    def forward(self, x):
        h     = self.LeakyReLU(self.FC_hidden(x))
        h     = self.bn1(self.LeakyReLU(self.FC_hidden2(h)))
        h     = self.LeakyReLU(self.FC_hidden3(h))
        h     = self.LeakyReLU(self.FC_output(h))
        next_time_ae = torch.sigmoid(self.FC_output1(h))
        return next_time_ae

class Med_pred_module(nn.Module):
    def __init__(self, latent_dim, hidden_dim, vocab_size, treatment_order, event_order):
        super(Med_pred_module, self).__init__()
        treatment_dim = vocab_size[treatment_order]
        med_dim = vocab_size[event_order]
        self.FC_hidden = nn.Linear(latent_dim+treatment_dim, latent_dim+10)
        self.FC_hidden2 = nn.Linear(latent_dim +10, hidden_dim)
        self.FC_hidden3 = nn.Linear(hidden_dim, hidden_dim)
        self.FC_output = nn.Linear(hidden_dim, med_dim)
        self.FC_output1 = nn.Linear(med_dim, med_dim)
        self.LeakyReLU= nn.ReLU()
        self.bn1 = nn.BatchNorm1d(hidden_dim)
    
    def forward(self, x):
        h     = self.LeakyReLU(self.FC_hidden(x))
        h     = self.bn1(self.LeakyReLU(self.FC_hidden2(h)))
        h     = self.LeakyReLU(self.FC_hidden3(h))
        h     = self.LeakyReLU(self.FC_output(h))
        next_time_med = torch.sigmoid(self.FC_output1(h))
        return next_time_med

class DotProductAttention(nn.Module):
    """
    Compute the dot products of the query with all values and apply a softmax function to obtain the weights on the values
    """
    def __init__(self):
        super(DotProductAttention, self).__init__()

    def forward(self, query, value):
        batch_size, input_size = query.size(0),  value.size(1)
        
        score = torch.bmm(query.unsqueeze(1), value.transpose(1, 2))
        attn = F.softmax(score, 2)  
     
        context = torch.bmm(attn, value)

        #print(context.shape)
        return context, attn

def loss_function(x, x_hat, mean, log_var, AE_out, AE_true):
    reproduction_loss = nn.functional.binary_cross_entropy(x_hat, x, reduction='sum')
    ae_loss = nn.functional.binary_cross_entropy(AE_out, AE_true, reduction='sum')
    KLD      = - 0.5 * torch.sum(1+ log_var - mean.pow(2) - log_var.exp())

    return (reproduction_loss + KLD) + ae_loss
class BuildModel(nn.Module):
    def __init__(self,
        hidden_dim,
        latent_dim,
        vocab_size,
        orders,
        event_type,
        device,
        epochs
        ) -> None:
        super().__init__()
        self.event_type = event_type # either medication or adverse event
        self.device = device
        self.ae_order=2
        self.med_order=1
        self. treatment_order = 0
        self.epochs = epochs

        if not isinstance(vocab_size, list): vocab_size = [vocab_size]

        if( event_type == 'medication'):
        #encoders for medications and adverse events
            self.Encoder_med = Encoder(vocab_size, self.med_order, hidden_dim, latent_dim)
            #decoders for medications and adverse events
            self.Decoder_med = Decoder(latent_dim, hidden_dim, vocab_size, self.med_order)
            #predictive modules for next time steps
            self.AE_pred = AE_pred_module(latent_dim, hidden_dim, vocab_size, self.treatment_order, self.ae_order)

        if(event_type == 'adverse events'):
            self.Encoder_ae = Encoder(vocab_size, self.ae_order, hidden_dim, latent_dim)
            self.Decoder_ae = Decoder(latent_dim, hidden_dim, vocab_size, self.ae_order)
            self.Med_pred = Med_pred_module(latent_dim, hidden_dim, vocab_size, self.treatment_order, self.med_order)

        #attention module
        self.Att = DotProductAttention()

    def reparameterization(self, mean, var):
        epsilon = torch.randn_like(var).to(self.device)        # sampling epsilon
        z = mean + var*epsilon                          # reparameterization trick
        return z

    def forward_med(self, med_x):#forward for medication reconstruction
        x = med_x[:,:, :-5]
        
        query = x[:, 0, :]
        keys = x[:, 1:, :]

        context, attn = self.Att(query, keys)

        med_mean, med_log_var = self.Encoder_med(context[:,0, :])
        z = self.reparameterization(med_mean, torch.exp(0.5 * med_log_var)) # takes exponential function (log var -> var)

        z_1= torch.cat((z, med_x[:,0, -5:]), 1)

        AE_out = self.AE_pred(z_1)

        med_x_hat  = self.Decoder_med(z)
        
        return med_x_hat, med_mean, med_log_var , AE_out

    def forward_ae(self, ae_x): #forward for adverse event reconstruction
        x = ae_x[:,:, :-5]
        
        
        query = x[:, 0, :]
        keys = x[:, 1:, :]
        context, _ = self.Att(query , keys )

        #print(context.shape, x.shape, query.shape, keys.shape)
        ae_mean, ae_log_var = self.Encoder_ae(context[:,0, :])
        z = self.reparameterization(ae_mean, torch.exp(0.5 * ae_log_var)) # takes exponential function (log var -> var)


        z_1= torch.cat((z, ae_x[:,0, -5:]), 1)
        Med_out = self.Med_pred(z_1)
        ae_x_hat = self.Decoder_ae(z)
        
        return ae_x_hat, ae_mean, ae_log_var , Med_out

    def forward(self, x): #last five columns are treatments
        if( self.event_type == 'medication'):
          x_hat, mean, log_var , out = self.forward_med(x)
        if (self.event_type == 'adverse events'):
          x_hat, mean, log_var , out = self.forward_ae( x)

        return x_hat, mean, log_var , out


class TWIN(SequenceSimulationBase):
    '''
    Implement a VAE based model for clinical trial patient digital twin simulation [1]_.
    
    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    order: list[str]
        The order of event types in each visits, e.g., ``['treatment', 'medication', 'adverse event']``.
        Visit = [treatment_events, medication_events, adverse_events], each event is a list of codes.

    event_type: str
        The type of event to be modeled, e.g., ``'medication'`` or ``'adverse event'``.

    max_visit: int
        Maximum number of visits.

    emb_size: int
        Embedding size for encoding input event codes.
        
    latent_dim: int
        Size of final latent dimension between the encoder and decoder

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    Notes
    -----
    .. [1] Trisha Das*, Zifeng Wang*, and Jimeng Sun. TWIN: Personalized Clinical Trial Digital Twin Generation. KDD'23.
    '''
    def __init__(self,
        vocab_size,
        order,
        event_type= 'medication',
        max_visit=13,
        emb_size=64,
        hidden_dim = 36,
        latent_dim=32,
        learning_rate=1e-3,
        batch_size=64,
        epochs=20,
        num_worker=0,
        device='cpu',# 'cuda:0',
        experiment_id='trial_simulation.sequence.twin',
        ):
        super().__init__(experiment_id)
        self.config = {
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'emb_size':emb_size,
            'hidden_dim': hidden_dim,
            'latent_dim':latent_dim,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':order,
            'event_type': event_type,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()
        
    def _build_model(self):
        self.model = BuildModel(
            hidden_dim = self.config['hidden_dim'],
            latent_dim=self.config['latent_dim'],
            vocab_size=self.config['vocab_size'],
            orders=self.config['orders'],
            event_type = self.config['event_type'],
            device=self.device,
            epochs = self.config['epochs']
            )
        self.model = self.model.to(self.device)

    def fit(self, train_data, out_dir):
        '''
        Train model with sequential patient records.

        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        self._input_data_check(train_data)
        df_train_data = self.SequencePatientBase_to_df(train_data)
        self._fit_model(df_train_data, out_dir)

    def next_step_df(self, data):
        if self.config['event_type'] == 'medication':
            ae_columns = [col for col in data if col.startswith('adverse events_')]
            for a in ae_columns:
                data[a[0:15]+'nxt_'+ a[15:]]= data[a].shift(-1)

            #create a new column by shifting up
            data['Visit_']= data['Visit'].shift(-1)
            data.iloc[len(data)-1,-1]=-1
            data = data[data['Visit_']-data['Visit']==1]
            data = data.drop(columns =['Visit_'])

            label_cols=[col for col in data if col.startswith('adverse events_nxt_')]
            y = data[label_cols]

            med_cols=[col for col in data if col.startswith('medication_')]
            treat_cols=[col for col in data if col.startswith('treatment_')]
            cols=med_cols+treat_cols
            X = data[cols]

        if self.config['event_type'] == 'adverse events':
            med_columns = [col for col in data if col.startswith('medication_')]
            for a in med_columns:
                data[a[0:11]+'nxt_'+ a[11:]]= data[a].shift(-1)

            #create a new column by shifting up
            data['Visit_']= data['Visit'].shift(-1)
            data.iloc[len(data)-1,-1]=-1
            data = data[data['Visit_']-data['Visit']==1]
            data = data.drop(columns =['Visit_'])

            label_cols=[col for col in data if col.startswith('medication_nxt_')]
            y = data[label_cols]

            ae_cols=[col for col in data if col.startswith('adverse events_')]
            treat_cols=[col for col in data if col.startswith('treatment_')]
            cols=ae_cols+treat_cols
            X = data[cols]
        return X, y

    def train(self, train_dl, device, optimizer, vocab_size, batch_size, model,out_dir):
      print("...Start training VAE...")
      print('--- event type: ', self.config['event_type'], '---')
      best_auc=0
      best_train_auc = 0

      for epoch in range(self.config['epochs']):
          overall_loss = 0
          for batch_idx, (x, y) in enumerate(train_dl):
                x = x.to(device)
                optimizer.zero_grad()
                n_cross_n = torch.matmul(x, x.T)
                top_5_index = torch.topk(n_cross_n, 4)
                ext_x=[]

                for i in range(x.shape[0]):
                    x_=[]
                    x_.append(x[i].tolist())
                    x_.extend(x[top_5_index.indices[i]].tolist())
                    ext_x.append(x_)

                ext_x= torch.as_tensor(ext_x)
                #print(ext_x[:,:, :-5])

                x_hat, mean, log_var , out = model(ext_x)

                loss= loss_function(x[:, :-5], x_hat, mean, log_var, out, y)

                overall_loss += loss.item() 

                loss.backward()
                optimizer.step()
                
          print("\tEpoch", epoch + 1, "complete!", "\tAverage Loss: ", overall_loss / (batch_idx*batch_size))

      self.save_model(output_dir=out_dir)
      print("Finish!!")

    def _fit_model(self, df, out_dir):
        X, y = self.next_step_df(df)
        train_dl= prepare_data(X, y, self.config['batch_size'])
        optimizer = Adam(self.model.parameters(), lr=self.config['learning_rate'])
        self.train(train_dl, self.device, optimizer, self.config['vocab_size'], self.config['batch_size'], self.model, out_dir)

    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatientBase), f'`trial_simulation.sequence` models require input training data in `SequencePatientBase`, find {type(inputs)} instead.'

    def SequencePatientBase_to_df(self, inputs):
        '''
        returns dataframe from SeqPatientBase
        '''
        #voc1 = inputs.voc
        inputs= inputs.visit
        column_names = ['People', 'Visit']
        for i in range(len(self.config['orders'])):
            for j in range(self.config['vocab_size'][i]):
              column_names.append(self.config['orders'][i]+'_'+str(j))

        df = pd.DataFrame(columns=column_names)
        for i in range(len(inputs)):#each patient
            for j in range(len(inputs[i])): #each visit
                binary_visit = [i, j]
                for k in range(len(inputs[i][j])): #k=3 types of events
                    event_binary= [0]*self.config['vocab_size'][k]
                    for l in inputs[i][j][k]: #multihot from dense
                        event_binary[l]=1
                    binary_visit.extend(event_binary)
                df.loc[len(df)] = binary_visit
        return df

    def df_to_SequencePatientBase(self, df, inputs):
        '''
        returns SeqPatientBase from df
        '''
        visits = []
        columns=[]
        for k in self.config['orders']:
            columns.extend([col for col in df if col.startswith('k')])
        for i in df.People.unique():
            sample=[]
            temp = df[df['People']==i]
            for index, row in temp.iterrows():
                visit=[]
                visit.append(np.nonzero(row[columns].to_list())[0].tolist())
                sample.append(visit)
            visits.append(sample)
            seqdata = SequencePatient(
                data={'v':visits, 'y': inputs.label, 'x':None},
                metadata={
                    'visit':seqdata.metadata['visit'],
                    'label':{'mode':'tensor'},
                    'voc':seqdata.metadata['voc'],
                    'max_visit':seqdata.metadata['max_visit'],
                }
            )
        return seqdata

    def predict(self, data ):
        self._input_data_check(data)
        df_data = self.SequencePatientBase_to_df(data)
        X, y = self.next_step_df(df_data)
        dl= prepare_data(X, y, self.config['batch_size'])
        self.model.eval()
        x_hats, ins= list(), list()
        for i, (x, y) in enumerate(dl):
            # evaluate the model on the test set
            n_cross_n = torch.matmul(x, x.T)
    
            #print(x.shape)
            top_5_index = torch.topk(n_cross_n, 4)
    
            ext_x=[]
            for j in range(x.shape[0]):
                x_=[]
                x_.append(x[j].tolist())
                x_.extend(x[top_5_index.indices[j]].tolist())
                ext_x.append(x_)
    
            ext_x= torch.as_tensor(ext_x)
    
            x_hat, mean, log_var, yhat= self.model(ext_x)
    
            inp = x.detach().numpy()
            x_hat = x_hat.detach().numpy()
            x_hat = x_hat.round()
            x_hats.append(x_hat)
            ins.append(inp)

        x_hats, ins=  vstack(x_hats), vstack(ins)
        return x_hats

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.
        
        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename, map_location=self.config['device'])
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config = config
        if self.config['event_type']=='medication':
            self.model.Encoder_med.load_state_dict(state_dict['encoder'])
            self.model.Decoder_med.load_state_dict(state_dict['decoder'])
            self.model.AE_pred.load_state_dict(state_dict['predictor'])
        if self.config['event_type']=='adverse events':
            self.model.Encoder_ae.load_state_dict(state_dict['encoder'])
            self.model.Decoder_ae.load_state_dict(state_dict['decoder'])
            self.model.Med_pred.load_state_dict(state_dict['predictor'])

    def _save_config(self, config, output_dir=None):
        temp_path = os.path.join(output_dir, self.config['event_type']+'_twin_config.json')
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def save_model(self, output_dir):
        '''
        Save the learned simulation model to the disk.
        Parameters
        ----------
        output_dir: str
            The dir to save the learned model.
        '''
        make_dir_if_not_exist(output_dir)
        self._save_config(config=self.config, output_dir=output_dir)
        if (self.config['event_type']=='medication'):
            self._save_checkpoint({
                    'encoder': self.model.Encoder_med.state_dict(),
                    'decoder': self.model.Decoder_med.state_dict(),
                    'predictor': self.model.AE_pred.state_dict()
                },output_dir=output_dir, filename='checkpoint_med.pth.tar')
        if (self.config['event_type']=='adverse events'):
            self._save_checkpoint({
                    'encoder': self.model.Encoder_ae.state_dict(),
                    'decoder': self.model.Decoder_ae.state_dict(),
                    'predictor': self.model.Med_pred.state_dict()
                },output_dir=output_dir, filename='checkpoint_ae.pth.tar')
        print('Save the trained model to:', output_dir)



import pdb
from copy import deepcopy

import transtab
import numpy as np
import pandas as pd
import torch

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import TabularIndivBase, IndivTabDataset

class BuildModel:
    def __new__(self, config):
        contrastive_pretrain = config.pop('contrastive_pretrain')
        if not contrastive_pretrain:
            clf = transtab.build_classifier(**config)
            collate_fn = None
        else:
            clf, collate_fn = transtab.build_contrastive_learner(**config)
        return clf, collate_fn

class TransTab(TabularIndivBase):
    '''
    Implement transtab model for tabular individual outcome
    prediction in clinical trials [1]_.

    Parameters
    ----------
    mode: str
        The task's objectives, in `binary`, `multiclass`. # TODO: `multilabel`, or `regression`
        Can be ignored if `contrastive_pretrain` is set True.
    
    categorical_columns: list 
        a list of categorical feature names.

    numerical_columns: list
        a list of numerical feature names.

    binary_columns: list
        a list of binary feature names, accept binary indicators like (yes,no); (true,false); (0,1).
    
    contrastive_pretrain: bool(default=False)
        whether or not take a contrastive pretraining. If set true,
        `num_class` will be ignored.

    num_class: int
        number of output classes to be predicted.

    hidden_dim: int
        the dimension of hidden embeddings.
    
    num_layer: int
        the number of transformer layers used in the encoder.
    
    num_attention_head: int
        the numebr of heads of multihead self-attention layer in the transformers.

    hidden_dropout_prob: float
        the dropout ratio in the transformer encoder.

    ffn_dim: int
        the dimension of feed-forward layer in the transformer layer.
    
    activation: str
        the name of used activation functions, support ``"relu"``, ``"gelu"``, ``"selu"``, ``"leakyrelu"``.

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        Target device to train the model, as `cuda:0` or `cpu`.

    experiment_id: str, optional (default='test')
        The name of current experiment. Decide the saved model checkpoint name.
    
    Notes
    -----
    .. [1] Wang, Z., & Sun, J. (2022). TransTab: Learning Transferable Tabular Transformers Across Tables. NeurIPS'22.
    '''
    def __init__(self, 
        mode=None,
        categorical_columns=None,
        numerical_columns=None,
        binary_columns=None,
        contrastive_pretrain=False,
        num_class=2,
        hidden_dim=128,
        num_layer=2,
        num_attention_head=8,
        hidden_dropout_prob=0,
        ffn_dim=256,
        activation='relu',
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test'):
        super().__init__(experiment_id)

        mode = mode.lower()
        if not contrastive_pretrain:
            assert mode in ['binary', 'multiclass', 'regression', 'multilabel'], 'Must specify `mode` for supervised classifcation.'

        self.config = {
            'categorical_columns':categorical_columns,
            'numerical_columns':numerical_columns,
            'binary_columns':binary_columns,
            'contrastive_pretrain':contrastive_pretrain,
            'num_class':num_class,
            'hidden_dim':hidden_dim,
            'num_layer':num_layer,
            'num_attention_head':num_attention_head,
            'hidden_dropout_prob':hidden_dropout_prob,
            'ffn_dim':ffn_dim,
            'activation':activation,
            'device':device,
            'mode':mode,
            'learning_rate':learning_rate,
            'weight_decay':weight_decay,
            'batch_size':batch_size,
            'epochs':epochs,
            'num_worker':num_worker,
        }
        self._save_config(self.config)
        self._build_model()
        self.device = device

    def fit(self, train_data, valid_data=None):
        '''Train TransTab model to predict patient outcome
        with tabular input data.

        Parameters
        ----------
        train_data: list[dict]
            a list of patient data, each patient is a dict of 
            {
            
            'x': TabularPatientBase or pd.DataFrame, 
            
            'y': pd.Series or np.ndarray
            
            }.

            TransTab can learn from multiple different tabular datasets.

        valid_data: dict
            Validation data during the training for early stopping. 
            valid_data = 
            
            {
            
            'x': TabularPatientBase or pd.DataFrame,

            'y': pd.Series or np.ndarray
            
            }
            
        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data=train_data, valid_data=valid_data)

    def predict(self, test_data):
        '''
        Make prediction probability based on the learned model.

        Parameters
        ----------
        test_data: TabularPatientBase or pd.DataFrame
            Contain all patient features.

        Returns
        -------
        ypred: np.ndarray or torch.Tensor

            - For binary classification, return shape (n, );
            - For multiclass classification, return shape (n, n_class).

        '''
        self._input_data_check(test_data)
        data = self._parse_input_data(test_data)
        ypred = transtab.predict(self.model, x_test=data)
        return ypred
        
    def save_model(self, output_dir=None):
        '''
        Save the learned transtab model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        self.model.save(output_dir)

    def load_model(self, checkpoint):
        '''
        Load the learned transtab model from the given checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.

            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        config_filename = check_model_config_file(checkpoint)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)        
        self.model.load(checkpoint)

    def update(self, config):
        '''Update the configuration of feature extractor's column map for *cat*, *num*, and *bin* cols.
        Or update the number of classes for the output classifier layer.

        Parameters
        ----------
        config: dict
            a dict of configurations: keys `cat:list`, `num:list`, `bin:list` are to specify the new column names;
            key `num_class:int` is to specify the number of classes for finetuning on a new dataset.

        '''
        self.model.update(config)

    def _build_model(self):
        config = deepcopy(self.config)
        self.model, self.collate_fn = BuildModel(config)

    def _fit_model(self, train_data, valid_data=None):
        train_data = self._parse_input_data(train_data)
        if valid_data is not None: valid_data = self._parse_input_data(valid_data)
        
        if self.config['mode'] == 'binary': eval_metric = 'auc'
        elif self.config['mode'] == 'multiclass': eval_metric = 'acc'

        transtab.train(
            self.model,
            train_data, 
            valset=valid_data, 
            eval_metric=eval_metric,
            output_dir=self.checkout_dir,
            collate_fn = self.collate_fn,
            **self.config
            )
        
    def _parse_input_data(self, inputs):
        def _check_input(input):
            if isinstance(input, pd.DataFrame):
                return input
            if isinstance(input, TabularPatientBase):
                return input.df

        if isinstance(inputs, list):
            data = []
            for input in inputs:
                x = _check_input(input['x'])
                data.append((x, input['y']))

        if isinstance(inputs, dict):
            data = (_check_input(inputs['x']), inputs['y'])

        if isinstance(inputs, pd.DataFrame):
            data = inputs
        
        if isinstance(inputs, TabularPatientBase):
            data = inputs.df

        return data

    def _input_data_check(self, inputs):
        '''
        Check the training / testing data fits the formats.
        Target to (1) check if inputs valid,
                    if not, give tips about the data problem.

        Parameters
        ----------
        inputs: [{
                'x': TabularPatientBase or pd.DataFrame,
                'y': pd.Series or np.ndarray
                },...]
                'x' contain all patient features; 'y' contain labels
                for each row.
        '''
        def _check_input(input):
            if isinstance(input, dict):
                assert 'x' in input, 'No input patient data found in inputs.'
                assert isinstance(input['x'], pd.DataFrame) or isinstance(input['x'], TabularPatientBase), 'Get unaccepted input data format, expect `pd.DataFrame` or `TabularPatientBase`, get {} instead.'.format(type(inputs['x']))
                if 'y' in input:
                    assert isinstance(input['y'], pd.Series) or isinstance(input['y'], np.ndarray)
                    assert not pd.isnull(input['y']).any(), 'Find NaN in input targets, please check.'
            if isinstance(input['x'], pd.DataFrame):
                assert not input['x'].isnull().values.any(), 'Find NaN in input dataframe, please check your input, or try to pass `TabularPatientBase` as inputs.'
            if isinstance(input['x'], TabularPatientBase):
                assert not input['x'].df.isnull().values.any(), 'Find NaN in input dataset, please check your input, or try to pass `TabularPatientBase` as inputs.'

        if isinstance(inputs, list):
            for input in inputs:
                _check_input(input)
        else:
            _check_input(inputs)



import pdb
import os
import itertools

import numpy as np
import torch
import random

def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.random.set_rng_state(torch.manual_seed(seed).get_state())

def admet_collate_fn(x):
    smiles_lst = [i[0] for i in x]
    label_vec = default_collate([int(i[1]) for i in x])  ### shape n, 
    return [smiles_lst, label_vec]

def smiles_txt_to_lst(text):
    """
        "['CN[C@H]1CC[C@@H](C2=CC(Cl)=C(Cl)C=C2)C2=CC=CC=C12', 'CNCCC=C1C2=CC=CC=C2CCC2=CC=CC=C12']" 
    """
    text = text[1:-1]
    lst = [i.strip()[1:-1] for i in text.split(',')]
    return lst

def icdcode_text_2_lst_of_lst(text):
    text = text[2:-2]
    lst_lst = []
    for i in text.split('", "'):
        i = i[1:-1]
        lst_lst.append([j.strip()[1:-1] for j in i.split(',')])
    return lst_lst

def list_text_2_list(text):
    text = text[2:-2]
    lst = [i.strip() for i in text.split("', '")]
    return lst

def flatten_stacked_list(lst):
    return list(itertools.chain(*lst))

def clean_protocol(protocol):
    try:
        protocol = protocol.lower()
    except:
        pdb.set_trace()
    protocol_split = protocol.split('\n')
    filter_out_empty_fn = lambda x: len(x.strip())>0
    strip_fn = lambda x: x.strip()
    protocol_split = list(filter(filter_out_empty_fn, protocol_split))	
    protocol_split = list(map(strip_fn, protocol_split))
    return protocol_split

def split_protocol(protocol):
    protocol_split = clean_protocol(protocol)
    inclusion_idx, exclusion_idx = len(protocol_split), len(protocol_split)	
    for idx, sentence in enumerate(protocol_split):
        if "inclusion" in sentence:
            inclusion_idx = idx
            break
    for idx, sentence in enumerate(protocol_split):
        if "exclusion" in sentence:
            exclusion_idx = idx 
            break 		
    if inclusion_idx + 1 < exclusion_idx + 1 < len(protocol_split):
        inclusion_criteria = protocol_split[inclusion_idx:exclusion_idx]
        exclusion_criteria = protocol_split[exclusion_idx:]
        if not (len(inclusion_criteria) > 0 and len(exclusion_criteria) > 0):
            print(len(inclusion_criteria), len(exclusion_criteria), len(protocol_split))
            exit()
        return inclusion_criteria, exclusion_criteria ## list, list 
    else:
        return protocol_split, []

def collect_trials_from_topic(topic):
    '''
    Collect all trials from a topic.

    Parameters
    ----------
    topic: Topic
        A `Topic` object.
    
    Returns
    -------
    trials: List[str]
        A list of `Trial`'s `nct_id`.
    '''
    trials = []
    for trial_step in topic.trials:
        ids = [trial.attr_dict['nctid'] for trial in trial_step]
        trials.extend(ids)
    return trials


import pdb
import os

import pandas as pd
import torch
from torch import nn
from torch.utils.data import DataLoader

from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)

from .base import TabularIndivBase, IndivTabDataset

class BuildModel(nn.Module):
    def __init__(self,
        input_dim,
        output_dim,
        hidden_dim=128,
        num_layer=2,
        ) -> None:
        super().__init__()
        if num_layer == 1:
            self.mlp = nn.ModuleList([nn.Linear(input_dim, output_dim)])
        else:
            self.mlp = nn.ModuleList([nn.Linear(input_dim, hidden_dim)])

        for _ in range(num_layer-2):
            self.mlp.append(nn.ReLU())
            self.mlp.append(nn.Linear(hidden_dim, hidden_dim))
        
        if num_layer > 1:
            self.mlp.append(nn.ReLU())
            self.mlp.append(nn.Linear(hidden_dim, output_dim))
    
    def forward(self, inputs):
        if isinstance(inputs, dict):
            h = inputs['x']
        elif isinstance(inputs, torch.Tensor):
            h = inputs
        else:
            raise ValueError('inputs should be dict or torch.Tensor')

        h = h.float()
        for layer in self.mlp:
            h = layer(h)
        return h

class MLP(TabularIndivBase):
    '''
    Implement multi-layer perceptron model for tabular individual outcome
    prediction in clinical trials.

    Parameters
    ----------
    input_dim: int
        Dimension of the input features.

    output_dim: int
        Dimension of the outputs. When doing classification, it equals to number of classes.

    mode: str
        The task's objectives, in `binary`, `multiclass`, `multilabel`, or `regression`

    hidden_dim: int
        Hidden dimensions of neural networks.

    num_layer: int
        Number of hidden layers.

    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weigth_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        Target device to train the model, as `cuda:0` or `cpu`.

    experiment_id: str, optional (default='test')
        The name of current experiment. Decide the saved model checkpoint name.
    '''
    def __init__(self,
        input_dim,
        output_dim,
        mode,
        hidden_dim=128,
        num_layer=2,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test'):
        super().__init__(experiment_id)

        mode = mode.lower()
        assert mode in ['binary', 'multiclass', 'regression', 'multilabel']

        self.config = {
            'input_dim':input_dim,
            'output_dim':output_dim,
            'hidden_dim':hidden_dim,
            'num_layer':num_layer,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'experiment_id':experiment_id,
            'model_name': 'MLP',
            'device':device,
            'mode':mode,
        }
        self._save_config(self.config)
        self.device = device

    def fit(self, train_data, valid_data=None):
        '''Train logistic regression model to predict patient outcome
        with tabular input data.

        Parameters
        ----------
        train_data: dict
            {
            'x': TabularPatientBase or pd.DataFrame,
            'y': pd.Series or np.ndarray
            }

            - 'x' contain all patient features; 
            - 'y' contain labels for each row.

        valid_data: same as `train_data`.
            Validation data during the training for early stopping.
        '''

        self._input_data_check(train_data)
        self._build_model()
        
        x_feat, y = self._parse_input_data(train_data)
        train_data={'x':x_feat, 'y':y}

        if valid_data is not None:
            x_feat_va, y_va = self._parse_input_data(valid_data)
            valid_data = {'x':x_feat_va, 'y':y_va}
                
        self._fit_model(train_data=train_data, valid_data=valid_data)
    
    def predict(self, test_data):
        '''
        Make prediction probability based on the learned model.

        Parameters
        ----------
        test_data: Dict or TabularPatientBase or pd.DataFrame or torch.Tensor
            {'x': TabularPatientBase or pd.DataFrame}
            
            'x' contain all patient features.

        Returns
        -------
        ypred: np.ndarray or torch.Tensor
            Prediction probability for each patient.

            - For binary classification, return shape (n, );
            - For multiclass classification, return shape (n, n_class).

        '''

        if isinstance(test_data, torch.Tensor):
            return self.model(test_data)
        
        x_feat, y = self._parse_input_data(test_data)
        test_data = {'x':x_feat, 'y': y}
        dataset = self._build_dataset(test_data)
        dataloader = DataLoader(dataset, 
            batch_size=self.config['batch_size'], 
            shuffle=False, 
            num_workers=self.config['num_worker'],
            pin_memory=True,
            )
        ypred = self._predict_on_dataloader(dataloader)
        return ypred

    def save_model(self, output_dir=None):
        '''
        Save the learned logistic regression model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model}, output_dir=output_dir)

    def load_model(self, checkpoint):
        '''
        Load model and the pre-encoded trial embeddings from the given
        checkpoint dir.

        Parameters
        ----------
        checkpoint: str
            The input dir that stores the pretrained model.
            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        self.model = state_dict['model']

    def _build_model(self):
        self.model = BuildModel(
            input_dim=self.config['input_dim'],
            output_dim=self.config['output_dim'],
            hidden_dim=self.config['hidden_dim'],
            )
        self.model.to(self.config['device'])



    







import pdb

import torch
from torch import nn

from pytrial.utils.check import (
    check_checkpoint_file, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceIndivBase
from ..trainer import IndivSeqTrainer


class RETAINLayer(nn.Module):
    """The separate callable RETAIN layer.
    Args:
        input_size: the embedding size of the input
        output_size: the embedding size of the output
        num_layers: the number of layers in the RNN
        dropout: dropout rate
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        dropout: float = 0.5,
        ):
        super(RETAINLayer, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.dropout = dropout
        self.dropout_layer = nn.Dropout(p=self.dropout)

        self.alpha_gru = nn.GRU(input_size, hidden_size, batch_first=True)
        self.beta_gru = nn.GRU(input_size, hidden_size, batch_first=True)

        self.alpha_li = nn.Linear(hidden_size, 1)
        self.beta_li = nn.Linear(hidden_size, hidden_size)

    def forward(self, inputs):
        """Using the sum of the embedding as the output of the transformer
        Args:
            x: [batch size, seq len, input_size]
            mask: [batch size, seq len]
        Returns:
            outputs [batch size, seq len, hidden_size]
        """
        x = inputs['x']
        mask = inputs['mask']

        # rnn will only apply dropout between layers
        x = self.dropout_layer(x)

        g, _ = self.alpha_gru(x)  # (patient, seq_len, hidden_size)
        h, _ = self.beta_gru(x)  # (patient, seq_len, hidden_size)

        # TOFIX: mask out the visit (by adding a large negative number 1e10)
        # however, it does not work better than not mask out
        attn_g = self.alpha_li(g)

        if mask is not None:
            attn_g = self._apply_mask_to_attention(attn_g, mask)

        attn_g = torch.softmax(attn_g, dim=1)  # (patient, seq_len, 1)
        attn_h = torch.tanh(self.beta_li(h))  # (patient, seq_len, hidden_size)
        c = attn_g * attn_h * x  # (patient, seq_len, hidden_size)
        # c = torch.sum(c, dim=1)  # (patient, hidden_size)
        return {'x':c, 'mask':mask}
    
    def _apply_mask_to_attention(self, attn_g, mask):
        mask_offset = torch.zeros(mask.shape, device=mask.device)
        mask_offset[mask==0] = -1e9
        mask_offset[mask==1] = 0
        attn_g += mask_offset.unsqueeze(-1)
        return attn_g


class BuildModel(nn.Module):
    def __init__(self,
        emb_size,
        hidden_dim,
        vocab_size,
        n_rnn_layer,
        output_dim,
        **kwargs,
        ) -> None:
        super().__init__()
        if not isinstance(vocab_size, list): vocab_size = [vocab_size]
        self.emb_size = emb_size
        self.vocab_size = vocab_size
        self.total_vocab_size = sum(vocab_size)
        self.embedding_matrix = nn.Linear(self.total_vocab_size, emb_size, bias=False)
        rnn_list = []
        for i in range(n_rnn_layer):
            if i == 0:
                rnn = RETAINLayer(
                        input_size=emb_size,
                        hidden_size=hidden_dim,
                        dropout=0.5,
                    )
            
            else:
                rnn = RETAINLayer(
                    input_size=hidden_dim,
                    hidden_size=hidden_dim,
                    dropout=0.5,
                )
            rnn_list.append(rnn)
        self.rnn = nn.Sequential(*rnn_list)
        self.predictor = nn.Linear(hidden_dim, output_dim)

    def forward(self, inputs):
        v = inputs['v']
        visit_emb = self.embedding_matrix(v) # bs, num_visit, emb_size

        if 'v_lengths' in inputs:
            v_len = inputs['v_lengths']
            mask = self._create_mask_from_visit_length(v_len=v_len)
            mask = mask.to(visit_emb.device)
        else:
            mask = None

        output = self.rnn({'x':visit_emb, 'mask':mask}) # bs, seq_len, hidden_size
        output = output['x']
        output = torch.sum(output, dim=1) # bs, emd_dim
        return self.predictor(output)

    def _create_mask_from_visit_length(self, v_len):
        mask = torch.zeros(len(v_len),max(v_len))
        for i in range(len(v_len)):
            mask[i, :v_len[i]] = 1
        return mask


class RETAIN(SequenceIndivBase):
    '''
    Implement RETAIN for longitudinal patient records predictive modeling [1]_.

    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    orders: list[str]
        A list of orders when treating inputs events. Should have the same shape of `vocab_size`.

    output_dim: int
        The dimension of the output.

        - If binary classification, output_dim=1;
        - If multiclass/multilabel classification, output_dim=n_class
        - If regression, output_dim=1.

    mode: str
        Prediction traget in ['binary','multiclass','multilabel','regression'].

    max_visit: int
        The maximum number of visits for input event codes.

    emb_size: int
        Embedding size for encoding input event codes.

    n_rnn_layer: int
        Number of RETAIN layers for encoding historical events.
    
    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.
    
    Notes
    -----
    .. [1] Choi, E., Bahadori, M. T., Sun, J., Kulas, J., Schuetz, A., & Stewart, W. (2016). Retain: An interpretable predictive model for healthcare using reverse time attention mechanism. Advances in neural information processing systems, 29.
    '''
    def __init__(self,
        vocab_size,
        orders,
        mode,
        output_dim=None,
        max_visit=None,
        emb_size=64,
        n_rnn_layer=2,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ):
        super().__init__(experiment_id, mode=mode, output_dim=output_dim)
        self.config = {
            'mode':self.mode,
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'emb_size':emb_size,
            'hidden_dim':emb_size,
            'n_rnn_layer':n_rnn_layer,
            'output_dim':self.output_dim,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':orders,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()

    def fit(self, train_data, valid_data=None):
        '''
        Train model with sequential patient records.
        
        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events; 'y' corresponds to labels.

        valid_data: SequencePatientBase
            A `SequencePatientBase` contains patient records used to make early stopping of the
            model.
        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data=valid_data)

    def predict(self, test_data):
        '''
        Predict patient outcomes using longitudinal trial patient sequences.
        
        Parameters
        ----------
        test_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.
        '''
        test_dataloader = self.get_test_dataloader(test_data)
        outputs = self._predict_on_dataloader(test_dataloader)
        return outputs['pred']

    def load_model(self, checkpoint):
        '''
        Load pretrained model from the disk.

        Parameters
        ----------
        checkpoint: str
            The input directory that stores the trained pytorch model and configuration.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        
        self.model.load_state_dict(state_dict['model'])

    def save_model(self, output_dir):
        '''
        Save the pretrained model to the disk.

        Parameters
        ----------
        output_dir: str
            The output directory that stores the trained pytorch model and configuration.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model.state_dict()}, output_dir=output_dir)

    def _build_model(self):
        self.model = BuildModel(
            emb_size=self.config['emb_size'],
            hidden_dim=self.config['hidden_dim'],
            n_rnn_layer=self.config['n_rnn_layer'],
            vocab_size=self.config['vocab_size'],
            output_dim=self.config['output_dim'],
            )
        self.model.to(self.device)

    def _fit_model(self, train_data, valid_data=None):
        test_metric_dict = {
            'binary': 'auc',
            'multiclass': 'acc',
            'regression': 'mse',
            'multilabel': 'f1', # take average of F1 scores
            }
        train_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model) for loss_model in loss_models]

        if self.config['mode'] == 'regression':
            less_is_better = True
        else:
            less_is_better = False

        trainer = IndivSeqTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,
            test_metric=test_metric_dict[self.config['mode']],
            less_is_better=less_is_better,
            )
        trainer.train(**self.config)


'''
Implement TVAE model for tabular simulation
prediction in clinical trials.
'''
import os
import warnings
import joblib
import torch
from torch.nn.functional import cross_entropy
from torch.utils.data import DataLoader, TensorDataset
from torch.optim import Adam
from ctgan import TVAESynthesizer as TVAESynthesizerBase
from ctgan.synthesizers.tvae import Encoder, Decoder

from .base import TabularSimulationBase
from .copula_gan import DataTransformer
from pytrial.data.patient_data import TabularPatientBase
from pytrial.utils.check import check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist

warnings.filterwarnings('ignore')

def _loss_function(recon_x, x, sigmas, mu, logvar, output_info, factor):
    st = 0
    loss = []
    for column_info in output_info:
        for span_info in column_info:
            if span_info.activation_fn != 'softmax':
                ed = st + span_info.dim
                std = sigmas[st]
                eq = x[:, st] - torch.tanh(recon_x[:, st])
                loss.append((eq ** 2 / 2 / (std ** 2)).sum())
                loss.append(torch.log(std) * x.size()[0])
                st = ed

            else:
                ed = st + span_info.dim
                loss.append(cross_entropy(
                    recon_x[:, st:ed], torch.argmax(x[:, st:ed], dim=-1), reduction='sum'))
                st = ed

    assert st == recon_x.size()[1]
    KLD = -0.5 * torch.sum(1 + logvar - mu**2 - logvar.exp())
    return sum(loss) * factor / x.size()[0], KLD / x.size()[0]

class TVAESynthesizer(TVAESynthesizerBase):
    def fit(self, train_data, discrete_columns=()):
        """Fit the TVAE Synthesizer models to the training data.

        Args:
            train_data (numpy.ndarray or pandas.DataFrame):
                Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
            discrete_columns (list-like):
                List of discrete columns to be used to generate the Conditional
                Vector. If ``train_data`` is a Numpy array, this list should
                contain the integer indices of the columns. Otherwise, if it is
                a ``pandas.DataFrame``, this list should contain the column names.
        """
        self.transformer = DataTransformer()
        self.transformer.fit(train_data, discrete_columns)
        train_data = self.transformer.transform(train_data)
        dataset = TensorDataset(torch.from_numpy(train_data.astype('float32')).to(self._device))
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True, drop_last=False)

        data_dim = self.transformer.output_dimensions
        encoder = Encoder(data_dim, self.compress_dims, self.embedding_dim).to(self._device)
        self.decoder = Decoder(self.embedding_dim, self.decompress_dims, data_dim).to(self._device)
        optimizerAE = Adam(
            list(encoder.parameters()) + list(self.decoder.parameters()),
            weight_decay=self.l2scale)

        for i in range(self.epochs):
            for id_, data in enumerate(loader):
                optimizerAE.zero_grad()
                real = data[0].to(self._device)
                mu, std, logvar = encoder(real)
                eps = torch.randn_like(std)
                emb = eps * std + mu
                rec, sigmas = self.decoder(emb)
                loss_1, loss_2 = _loss_function(
                    rec, real, sigmas, mu, logvar,
                    self.transformer.output_info_list, self.loss_factor
                )
                loss = loss_1 + loss_2
                loss.backward()
                optimizerAE.step()
                self.decoder.sigma.data.clamp_(0.01, 1.0)


class BuildModel:
    def __new__(self, config) -> TVAESynthesizer:
        model = TVAESynthesizer(
            embedding_dim=config['embedding_dim'],
            compress_dims=config['compress_dims'],
            decompress_dims=config['decompress_dims'],
            l2scale=config['l2scale'],
            batch_size=config['batch_size'],
            epochs=config['epochs'],
            loss_factor=config['loss_factor'],
            cuda=config['cuda'],
            )

        return model


class TVAE(TabularSimulationBase):
    '''
    Implement TVAE model for tabular patient data simulation [1]_.

    Parameters
    ----------
    embedding_dim: int
        Size of the random sample passed to the Generator. Defaults to 128.

    compress_dims: tuple or list[int]
        Size of each hidden layer in the encoder. Defaults to (128, 128).
    
    decompress_dims: tuple or list[int]
       Size of each hidden layer in the decoder. Defaults to (128, 128).
    
    l2scale: int
        Regularization term. Defaults to 1e-5.
    
    batch_size: int
        Number of data samples to process in each step.
    
    epochs: int
        Number of training epochs. Defaults to 300.
    
    loss_factor: int
        Multiplier for the reconstruction error. Defaults to 2.
    
    cuda: bool or str
        - If ``True``, use CUDA. If a ``str``, use the indicated device.
        - If ``False``, do not use cuda at all.

    experiment_id: str, optional
        The name of current experiment. Decide the saved model checkpoint name.

    Notes
    -----
    .. [1] Xu, L., Skoularidou, M., Cuesta-Infante, A., & Veeramachaneni, K. (2019). Modeling tabular data using conditional gan. Advances in Neural Information Processing Systems, 32.
    '''
    def __init__(
        self,
        embedding_dim=128,
        compress_dims=(128, 128),
        decompress_dims=(128, 128),
        l2scale=1e-5,
        batch_size=500,
        epochs=50,
        loss_factor=2,
        cuda=False,
        experiment_id='trial_simulation.tabular.tvae',
    ) -> None:
        super().__init__(experiment_id=experiment_id)
        self.config = {
            'embedding_dim' : embedding_dim,
            'compress_dims' : compress_dims,
            'decompress_dims' : decompress_dims,
            'l2scale' : l2scale,
            'batch_size': batch_size,
            'epochs': epochs,
            'loss_factor' : loss_factor,
            'cuda' : cuda,
            'experiment_id': experiment_id,
            'model_name': 'tvae',
        }
        self._save_config(self.config)

    def fit(self, train_data):
        '''
        Train TVAE model to simulate patient data
        with tabular input data.

        Parameters
        ----------
        train_data: TabularPatientBase
            The training data for TVAE model.
        '''
        self._input_data_check(train_data)
        self._build_model()
        if isinstance(train_data, TabularPatientBase): # transform=True
            dataset = train_data.df
        if isinstance(train_data, dict): 
            dataset = TabularPatientBase(train_data, transform=True)
            dataset = dataset.df

        self.metadata = train_data.metadata
        self.raw_dataset = train_data

        dataset = self.raw_dataset.reverse_transform() # transform back

        categoricals = []
        fields_before_transform = self.metadata['sdtypes']
        for field in dataset.columns:
            field_name = field.replace('.value', '')
            if field_name in fields_before_transform:
                meta = fields_before_transform[field_name]
                if meta == 'categorical':
                    categoricals.append(field)

        self._fit_model(dataset, categoricals) 

    def predict(self, n=200):
        '''
        simulate a new tabular data with n.

        Parameters
        ----------
        n: int
            The number of new data to simulate.

        Returns
        -------
        ypred: TanularPatientBase
            A new tabular data simulated by the model
        '''
        ypred = self.model.sample(n) # build df
        return ypred # output: dataset, same as the input dataset not transform back

    def save_model(self, output_dir=None):
        '''
        Save the learned TVAE model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The dir to save the learned model.
            If set None, will save model to `self.checkout_dir`.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir

        self._save_config(self.config, output_dir=output_dir)
        ckpt_path = os.path.join(output_dir, 'tvae.model')
        joblib.dump(self.model, ckpt_path)

    def load_model(self, checkpoint=None):
        '''
        Load the learned TVAE model from the disk.

        Parameters
        ----------
        checkpoint: str or None
            The path to the checkpoint file.
            
            - If a directory, the only checkpoint file `.model` will be loaded.
            - If a filepath, will load from this file;
            - If None, will load from `self.checkout_dir`.
        '''
        if checkpoint is None:
            checkpoint = self.checkout_dir

        checkpoint_filename = check_checkpoint_file(checkpoint, suffix='model')
        config_filename = check_model_config_file(checkpoint)
        self.model = joblib.load(checkpoint_filename)
        self.config = self._load_config(config_filename)

    def _build_model(self):
        self.model = BuildModel(self.config)

    def _fit_model(self, data, discrete_columns):
        self.model.fit(data, discrete_columns=discrete_columns)



'''
Implement Doc2Vec model for clinical trial similarity search.
'''
import pdb
import os
import warnings
import json

import pandas as pd
import numpy as np
import torch
from torch import nn
import gensim
from gensim.models.doc2vec import Doc2Vec as doc2vec_model
from gensim.models.doc2vec import TaggedDocument

from pytrial.utils.check import check_checkpoint_file, check_model_config_file
from .base import TrialSearchBase

warnings.filterwarnings('ignore')

class BuildModel(nn.Module):
    def __init__(self,
        train_corpus,
        emb_dim=128,
        epochs=10,
        window=5,
        min_count=None,
        max_vocab_size=None,
        num_workers=None,
        ) -> None:
        super().__init__()

        # build gensim.Doc2vec model
        self.model = doc2vec_model(
            documents=train_corpus,
            vector_size=emb_dim,
            min_count=min_count,
            window=window,
            max_vocab_size=max_vocab_size,
            epochs=epochs,
            workers=num_workers,
            )

    def forward(self, x):
        x_w = pd.Series(x).apply(lambda x_: gensim.utils.simple_preprocess(x_))
        embs = x_w.apply(lambda x: self.model.infer_vector(x)[None])
        embs = np.concatenate(embs.values, 0)
        return embs

class Doc2Vec(TrialSearchBase):
    '''
    Implement the Doc2Vec model for trial document similarity search.

    Parameters
    ----------
    emb_dim: int, optional (default=128)
        Dimensionality of the embedding vectors.

    epochs: int, optional (default=10)
        Number of iterations (epochs) over the corpus. Defaults to 10 for Doc2Vec.

    window: int, optional (default=5)
        The maximum distance between the current and predicted word within a sentence.

    min_count: int, optional (default=5)
        Ignores all words with total frequency lower than this.

    max_vocab_size: int, optional
        Limits the RAM during vocabulary building; if there are more unique words than this, then prune the infrequent ones.
        Every 10 million word types need about 1GB of RAM. Set to None for no limit.

    num_workers: int, optional (default=4)
        Use these many worker threads to train the model (=faster training with multicore machines).

    experiment_id: str, optional (default='test')
        The name of current experiment.
    '''
    def __init__(self,
        emb_dim=128,
        epochs=10,
        window=5,
        min_count=5,
        max_vocab_size=None,
        num_workers=4,
        experiment_id='test',
        ) -> None:
        super().__init__(experiment_id=experiment_id)
        self.emb_dim=emb_dim
        self.epochs=epochs
        self.window=window
        self.min_count=min_count
        self.max_vocab_size=max_vocab_size
        self.num_workers=num_workers

    def fit(self, train_data, valid_data=None):
        '''Train the doc2vec model to get document embeddings for trial search.

        Parameters
        ----------
        train_data: dict

            Training corpus for the model. 
            
            - `x`: a dataframe of trial documents.
            
            - `fields`: optional, the fields of documents to use for training. If not given, the model uses all fields in `x`.

            - `tag`: optional, the field in `x` that serves as unique identifiers. Typically it is the `nct_id` of each trial. If not given, the model takes integer tags.

            train_data =

            {
            
            'x': pd.DataFrame,

            'fields': list[str],
        
            'tag': str,

            }


        valid_data: Ignored.
            Not used, present here for API consistency by convention.
        '''
        self._input_data_check(train_data)
        self._build_model(train_data)
        self._save_checkpoint(state=self.model, epoch_id=self.epochs)

    def encode(self, inputs):
        '''
        Encode input documents and output the document embeddings.

        Parameters
        ----------
        inputs: dict
            The documents which are to be encoded.
            `x`: a dataframe of trial documents.
            `fields`: the list of columns to be used in `x`.

            inputs = 

            {

            'x': pd.DataFrame,

            'fields': list[str],

            }

        Return
        ------
        embs: np.ndarray
            The encoded trial document embeddings.
        '''
        self._input_data_check(inputs)
        df = inputs['x']
        if 'fields' in inputs: fields = inputs['fields']
        else: fields = None
        df_raw = self._process_dataframe(df, fields)
        embs = self.model(df_raw['text'].tolist())
        return embs

    def predict(self, test_data, top_k=10):
        '''
        Take the input document, find the most similar documents in the training corpus.

        Parameters
        ----------
        test_data: dict
            Trial docs to be predicted. `x`: a dataframe of trial documents.
            `fields`: optional, the fields of documents to use for training. If not given,
            the model uses all fields in `x`.

            test_data =

            {

            'x': pd.DataFrame,

            'fields': list[str],

            }

        top_k: int, optional (default=10)
            The number of top similar documents to be retrieved.

        Return
        ------
        pred: list[list[tuple[str, float]]]
            The sequence of `('key': similarities)` for the input test documents for each input trial.
        '''
        self._input_data_check(test_data)

        # encode doc into embs
        embs = self.encode(test_data)

        # rank for each test document
        rank_list = []
        for emb in embs:
            ranks = self.model.model.dv.most_similar([emb], topn=top_k)
            rank_list.append(ranks)
        return rank_list

    def load_model(self, checkpoint):
        '''
        Load the pretrained model from disk.

        Parameters
        ----------
        checkpoint: str
            The path to the pretrained model.

            - If a directory, the only checkpoint file `*.pth.tar` will be loaded.
            - If a filepath, will load from this file.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        self.model = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_model_config(config_filename)
            self.config.update(config)

    def save_model(self, output_dir):
        '''
        Save the trained model.

        Parameters
        ----------
        output_dir: str
            The output directory to save. Checkpoint is saved with name `checkpoint.pth.tar` by default.
        '''
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        self._save_checkpoint(self.model, output_dir=output_dir)
        self._save_model_config(self.config, output_dir=output_dir)

    def __getitem__(self, tag):
        '''
        Get the embeddings of documents by the trial tags.

        Parameters
        ----------
        tag: str, int, list[str], list[int]
            The tag (or tags) to be looked up in the model.

        Returns
        -------
        emb: np.ndarry
            The embeddings of each document.
        '''
        return self.model.model[tag]

    def _build_model(self, train_data):
        # process raw pd.DataFrame inputs to the doc2vec tagged document inputs
        df = train_data['x']
        if 'fields' in train_data:
            fields = train_data['fields']
        else:
            fields = None
        df_raw = self._process_dataframe(df, fields)

        df_raw['text'] = df_raw['text'].apply(lambda x: gensim.utils.simple_preprocess(x))

        if 'tag' in train_data:
            tag = df[train_data['tag']]
            df_raw = pd.concat([tag, df_raw], axis=1)
            df_tagged_doc = df_raw.apply(lambda x: TaggedDocument(x.values[1], [x.values[0]]),axis=1)
        else:
            df_tagged_doc = df_raw.reset_index().apply(lambda x: TaggedDocument(x.values[1], [x.name]),axis=1)

        train_corpus = df_tagged_doc.tolist()

        config = {
            'emb_dim': self.emb_dim,
            'epochs': self.epochs,
            'window': self.window,
            'min_count': self.min_count,
            'max_vocab_size': self.max_vocab_size,
            'num_workers': self.num_workers,
        }
        self.model = BuildModel(train_corpus=train_corpus, **config)
        self.config = config
        self._save_model_config(self.config)



import pdb
from typing import Optional, List, Tuple, Dict, Union, Any

import torch
from torch import nn
import numpy as np
from transformers import Trainer
from transformers.trainer_pt_utils import (
    nested_detach, nested_concat, nested_truncate, nested_numpify, find_batch_size
)
from transformers.trainer_utils import has_length, denumpify_detensorize, EvalLoopOutput, EvalPrediction
from transformers.trainer_pt_utils import IterableDatasetShard
from torch.utils.data import DataLoader

class PromptEHRTrainer(Trainer):
    def __init__(self,
        model= None,
        args = None,
        data_collator=None,
        train_dataset=None,
        eval_dataset=None,
        val_data_collator=None,
        ):
        super().__init__(model, args, data_collator, train_dataset, eval_dataset)
        self.val_data_collator = val_data_collator if val_data_collator is not None else self.data_collator

    def get_train_dataloader(self) -> DataLoader:
        """
        Returns the training [`~torch.utils.data.DataLoader`].
        Will use no sampler if `train_dataset` does not implement `__len__`, a random sampler (adapted to distributed
        training if necessary) otherwise.
        Subclass and override this method if you want to inject some custom behavior.
        """
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")

        train_dataset = self.train_dataset
        data_collator = self.data_collator
        train_sampler = self._get_train_sampler()

        return DataLoader(
            train_dataset,
            batch_size=self._train_batch_size,
            sampler=train_sampler,
            collate_fn=data_collator,
            drop_last=self.args.dataloader_drop_last,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
        )

    def get_eval_dataloader(self, eval_dataset, code_type):
        """
        Returns the evaluation [`~torch.utils.data.DataLoader`].
        Subclass and override this method if you want to inject some custom behavior.
        Args:
            eval_dataset (`torch.utils.data.Dataset`, *optional*):
                If provided, will override `self.eval_dataset`. If it is an `datasets.Dataset`, columns not accepted by
                the `model.forward()` method are automatically removed. It must implement `__len__`.
        """
        if eval_dataset is None and self.eval_dataset is None:
            raise ValueError("Trainer: evaluation requires an eval_dataset.")
        eval_dataset = eval_dataset if eval_dataset is not None else self.eval_dataset
        self.val_data_collator.set_eval_code_type(code_type) # set evaluation for this code
        return DataLoader(
            eval_dataset,
            batch_size=self.args.eval_batch_size,
            collate_fn=self.val_data_collator,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
            shuffle=False,
            drop_last=False,
        )

    def evaluate(
        self,
        eval_dataset=None,
        ignore_keys=['encoder_last_hidden_state', 'past_key_values'],
        metric_key_prefix: str = "eval",
        ):
        """
        Run evaluation and returns metrics.

        The calling script will be responsible for providing a method to compute metrics, as they are task-dependent
        (pass it to the init :obj:`compute_metrics` argument).

        You can also subclass and override this method to inject custom behavior.

        Args:
            eval_dataset (:obj:`Dataset`, `optional`):
                Pass a dataset if you wish to override :obj:`self.eval_dataset`. If it is an :obj:`datasets.Dataset`,
                columns not accepted by the ``model.forward()`` method are automatically removed. It must implement the
                :obj:`__len__` method.
            ignore_keys (:obj:`Lst[str]`, `optional`):
                A list of keys in the output of your model (if it is a dictionary) that should be ignored when
                gathering predictions.
            metric_key_prefix (:obj:`str`, `optional`, defaults to :obj:`"eval"`):
                An optional prefix to be used as the metrics key prefix. For example the metrics "bleu" will be named
                "eval_bleu" if the prefix is "eval" (default)

        Returns:
            A dictionary containing the evaluation loss and the potential metrics computed from the predictions. The
            dictionary also contains the epoch number which comes from the training state.
        """
        # set evaluation mode
        self.model.eval()

        # memory metrics - must set up as early as possible
        self._memory_tracker.start()
        eval_loop = self.evaluation_loop
        code_type_list = self.data_collator.__code_type_list__
        output_metrics = {}
        for code_type in code_type_list:
            eval_dataloader = self.get_eval_dataloader(eval_dataset, code_type=code_type) # change the data collator
            output = eval_loop(
                eval_dataloader,
                description="Evaluation",
                # No point gathering the predictions if there are no metrics, otherwise we defer to
                # self.args.prediction_loss_only
                prediction_loss_only=False,
                ignore_keys=ignore_keys,
                metric_key_prefix=metric_key_prefix,
            )
            output_metrics[f'{metric_key_prefix}_ppl_{code_type}'] = output.metrics['eval_loss']

        output.metrics.update(output_metrics)
        output.metrics.pop('eval_loss')
        self.log(output.metrics)
        self.control = self.callback_handler.on_evaluate(self.args, self.state, self.control, output.metrics)
        self._memory_tracker.stop_and_update_metrics(output.metrics)

        # set training mode
        self.model.train()
        
        return output.metrics

    def evaluation_loop(
        self,
        dataloader: DataLoader,
        description: str,
        prediction_loss_only: Optional[bool] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> EvalLoopOutput:
        """
        Prediction/evaluation loop, shared by `Trainer.evaluate()` and `Trainer.predict()`.

        Works both with or without labels.
        """
        args = self.args
        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else args.prediction_loss_only
        model = self._wrap_model(self.model, training=False)

        # if full fp16 or bf16 eval is wanted and this ``evaluation`` or ``predict`` isn't called
        # while ``train`` is running, cast it to the right dtype first and then put on device
        if not self.is_in_train:
            if args.fp16_full_eval:
                model = model.to(dtype=torch.float16, device=args.device)
            elif args.bf16_full_eval:
                model = model.to(dtype=torch.bfloat16, device=args.device)

        batch_size = self.args.eval_batch_size

        print(f"***** Running {description} *****")
        if has_length(dataloader):
            print(f"  Num examples = {self.num_examples(dataloader)}")
        else:
            print("  Num examples: Unknown")
        print(f"  Batch size = {batch_size}")

        model.eval()

        self.callback_handler.eval_dataloader = dataloader
        # Do this before wrapping.
        eval_dataset = getattr(dataloader, "dataset", None)

        if args.past_index >= 0:
            self._past = None

        # Initialize containers
        # losses/preds/labels on GPU/TPU (accumulated for eval_accumulation_steps)
        losses_host = None
        preds_host = None
        labels_host = None
        inputs_host = None

        # losses/preds/labels on CPU (final containers)
        all_losses = None
        all_preds = None
        all_labels = None
        all_inputs = None
        # Will be useful when we have an iterable dataset so don't know its length.

        observed_num_examples = 0
        # Main evaluation loop
        for step, inputs in enumerate(dataloader):
            # Update the observed num examples
            observed_batch_size = find_batch_size(inputs)
            if observed_batch_size is not None:
                observed_num_examples += observed_batch_size
                # For batch samplers, batch_size is not known by the dataloader in advance.
                if batch_size is None:
                    batch_size = observed_batch_size

            # Prediction step
            loss, logits, labels = self.prediction_step(model, inputs, prediction_loss_only, ignore_keys=ignore_keys)
            inputs_decode = inputs["input_ids"] if args.include_inputs_for_metrics else None

            # Update containers on host
            if loss is not None:
                # if len(loss) == 1:
                #     # if output the mean batch loss
                #     losses = self._nested_gather(loss.repeat(batch_size))
                # else:
                #     # if output the raw batch loss sequence
                #     losses = loss
                losses = loss
                losses_host = losses if losses_host is None else torch.cat((losses_host, losses), dim=0)
            if labels is not None:
                labels = self._pad_across_processes(labels)
                labels = self._nested_gather(labels)
                labels_host = labels if labels_host is None else nested_concat(labels_host, labels, padding_index=-100)
            if inputs_decode is not None:
                inputs_decode = self._pad_across_processes(inputs_decode)
                inputs_decode = self._nested_gather(inputs_decode)
                inputs_host = (
                    inputs_decode
                    if inputs_host is None
                    else nested_concat(inputs_host, inputs_decode, padding_index=-100)
                )
            if logits is not None:
                logits = self._pad_across_processes(logits)
                logits = self._nested_gather(logits)
                if self.preprocess_logits_for_metrics is not None:
                    logits = self.preprocess_logits_for_metrics(logits, labels)
                preds_host = logits if preds_host is None else nested_concat(preds_host, logits, padding_index=-100)
            self.control = self.callback_handler.on_prediction_step(args, self.state, self.control)

            # Gather all tensors and put them back on the CPU if we have done enough accumulation steps.
            if args.eval_accumulation_steps is not None and (step + 1) % args.eval_accumulation_steps == 0:
                if losses_host is not None:
                    losses = nested_numpify(losses_host)
                    all_losses = losses if all_losses is None else np.concatenate((all_losses, losses), axis=0)
                if preds_host is not None:
                    logits = nested_numpify(preds_host)
                    all_preds = logits if all_preds is None else nested_concat(all_preds, logits, padding_index=-100)
                if inputs_host is not None:
                    inputs_decode = nested_numpify(inputs_host)
                    all_inputs = (
                        inputs_decode
                        if all_inputs is None
                        else nested_concat(all_inputs, inputs_decode, padding_index=-100)
                    )
                if labels_host is not None:
                    labels = nested_numpify(labels_host)
                    all_labels = (
                        labels if all_labels is None else nested_concat(all_labels, labels, padding_index=-100)
                    )

                # Set back to None to begin a new accumulation
                losses_host, preds_host, inputs_host, labels_host = None, None, None, None

        if args.past_index and hasattr(self, "_past"):
            # Clean the state at the end of the evaluation loop
            delattr(self, "_past")

        # Gather all remaining tensors and put them back on the CPU
        if losses_host is not None:
            losses = nested_numpify(losses_host)
            all_losses = losses if all_losses is None else np.concatenate((all_losses, losses), axis=0)
        if preds_host is not None:
            logits = nested_numpify(preds_host)
            all_preds = logits if all_preds is None else nested_concat(all_preds, logits, padding_index=-100)
        if inputs_host is not None:
            inputs_decode = nested_numpify(inputs_host)
            all_inputs = (
                inputs_decode if all_inputs is None else nested_concat(all_inputs, inputs_decode, padding_index=-100)
            )
        if labels_host is not None:
            labels = nested_numpify(labels_host)
            all_labels = labels if all_labels is None else nested_concat(all_labels, labels, padding_index=-100)

        # Number of samples
        if has_length(eval_dataset):
            num_samples = len(eval_dataset)
        # The instance check is weird and does not actually check for the type, but whether the dataset has the right
        # methods. Therefore we need to make sure it also has the attribute.
        elif isinstance(eval_dataset, IterableDatasetShard) and hasattr(eval_dataset, "num_examples"):
            num_samples = eval_dataset.num_examples
        else:
            if has_length(dataloader):
                num_samples = self.num_examples(dataloader)
            else:  # both len(dataloader.dataset) and len(dataloader) fail
                num_samples = observed_num_examples

        # Number of losses has been rounded to a multiple of batch_size and in a distributed training, the number of
        # samplers has been rounded to a multiple of batch_size, so we truncate.
        # if all_losses is not None:
        #     all_losses = all_losses[:num_samples]

        if all_preds is not None:
            all_preds = nested_truncate(all_preds, num_samples)
        if all_labels is not None:
            all_labels = nested_truncate(all_labels, num_samples)
        if all_inputs is not None:
            all_inputs = nested_truncate(all_inputs, num_samples)

        # Metrics!
        if self.compute_metrics is not None and all_preds is not None and all_labels is not None:
            if args.include_inputs_for_metrics:
                metrics = self.compute_metrics(
                    EvalPrediction(predictions=all_preds, label_ids=all_labels, inputs=all_inputs)
                )
            else:
                metrics = self.compute_metrics(EvalPrediction(predictions=all_preds, label_ids=all_labels))
        else:
            metrics = {}

        # To be JSON-serializable, we need to remove numpy types or zero-d tensors
        metrics = denumpify_detensorize(metrics)

        
        if all_losses is not None:
            if np.isnan(all_losses).any():
                # deal with nan
                print('[warning] find NaN in computed ppl, replace with 1e8, may cause incorrect ppl results.')
                all_losses = np.nan_to_num(all_losses, nan=1e8)

            # take median instead of mean for the sake of stability
            metrics[f"{metric_key_prefix}_loss"] = np.median(all_losses).item()

        # Prefix all keys with metric_key_prefix + '_'
        for key in list(metrics.keys()):
            if not key.startswith(f"{metric_key_prefix}_"):
                metrics[f"{metric_key_prefix}_{key}"] = metrics.pop(key)

        return EvalLoopOutput(predictions=all_preds, label_ids=all_labels, metrics=metrics, num_samples=num_samples)

    def prediction_step(
        self,
        model: nn.Module,
        inputs: Dict[str, Union[torch.Tensor, Any]],
        prediction_loss_only: bool,
        ignore_keys: Optional[List[str]] = None,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]:
        """
        Perform an evaluation step on `model` using `inputs`.
        Subclass and override to inject custom behavior.
        Args:
            model (`nn.Module`):
                The model to evaluate.
            inputs (`Dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.
                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument `labels`. Check your model's documentation for all accepted arguments.
            prediction_loss_only (`bool`):
                Whether or not to return the loss only.
            ignore_keys (`Lst[str]`, *optional*):
                A list of keys in the output of your model (if it is a dictionary) that should be ignored when
                gathering predictions.
        Return:
            Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]: A tuple with the loss,
            logits and labels (each being optional).
        """
        has_labels = all(inputs.get(k) is not None for k in self.label_names)
        inputs = self._prepare_inputs(inputs)
        if ignore_keys is None:
            if hasattr(self.model, "config"):
                ignore_keys = getattr(self.model.config, "keys_to_ignore_at_inference", [])
            else:
                ignore_keys = []

        # labels may be popped when computing the loss (label smoothing for instance) so we grab them first.
        if has_labels:
            labels = nested_detach(tuple(inputs.get(name) for name in self.label_names))
            if len(labels) == 1:
                labels = labels[0]
        else:
            labels = None
        
        with torch.no_grad():
            if has_labels:
                with self.autocast_smart_context_manager():
                    loss, outputs = self.compute_loss(model, inputs, return_outputs=True, return_perplexity=True)

                if loss is not None:
                    if not isinstance(loss, torch.Tensor):
                        # if more than one device is used
                        raise ValueError("In prediction phase only one GPU should be used, expected `torch.Tensor` loss, get `{}` instead.".format(type(loss)))
                    else:
                        loss = loss.item() # return ppl is a 0-d tensor
                    loss = torch.tensor([loss])

                if isinstance(outputs, dict):
                    logits = tuple(v for k, v in outputs.items() if k not in ignore_keys + ["loss"])
                else:
                    logits = outputs[1:]
            else:
                loss = None
                with self.autocast_smart_context_manager():
                    outputs = model(**inputs)
                if isinstance(outputs, dict):
                    logits = tuple(v for k, v in outputs.items() if k not in ignore_keys)
                else:
                    logits = outputs
                # TODO: this needs to be fixed and made cleaner later.
                if self.args.past_index >= 0:
                    self._past = outputs[self.args.past_index - 1]

        if prediction_loss_only:
            return (loss, None, None)

        logits = nested_detach(logits)
        if len(logits) == 1:
            logits = logits[0]

        return (loss, logits, labels)

    def compute_loss(self, model, inputs, return_outputs=False, return_perplexity=False):
        """
        How the loss is computed by Trainer. By default, all models return the loss in the first element.
        Subclass and override for custom behavior.

        Parameters
        ----------
        return_perplexity: whether or not return the computed perplexity instead of returning the loss.

        """
        outputs = model(**inputs)
        if return_perplexity:
            loss = outputs['perplexity']
        else:
            loss = outputs["loss"] if isinstance(outputs, dict) else outputs[0]
        return (loss, outputs) if return_outputs else loss

    def _prepare_inputs(self, data):
        """
        Prepares one :obj:`data` before feeding it to the model, be it a tensor or a nested list/dictionary of tensors.
        """
        if isinstance(data, dict):
            return type(data)(**{k: self._prepare_input(v) for k, v in data.items()})
        elif isinstance(data, (tuple, list)):
            return type(data)(self._prepare_input(v) for v in data)
        elif isinstance(data, torch.Tensor):
            kwargs = dict(device=self.args.device)
            if self.deepspeed and data.dtype != torch.int64:
                # NLP models inputs are int64 and those get adjusted to the right dtype of the
                # embedding. Other models such as wav2vec2's inputs are already float and thus
                # may need special handling to match the dtypes of the model
                kwargs.update(dict(dtype=self.args.hf_deepspeed_config.dtype()))
            return data.to(**kwargs)
        return data



import pdb

import torch
from torch import nn

class XentLoss(nn.Module):
    '''
    The basic xent loss
    for multi-class classification.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.CrossEntropyLoss()
    
    def forward(self, inputs):
        logits = self.model(inputs)
        loss = self.xent_loss(logits, inputs['y'])
        return {'loss_value':loss}

class BinaryXentLoss(nn.Module):
    '''
    The basic binary xent loss
    for binary classification.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.BCEWithLogitsLoss()

    def forward(self, inputs):
        logits = self.model(inputs)
        y = inputs['y'].float()
        if y.shape != logits.shape:
            logits = logits.reshape(y.shape)
        loss = self.xent_loss(logits, y)
        return {'loss_value':loss}

class MultilabelBinaryXentLoss(nn.Module):
    '''
    The basic binary xent loss
    for multilabel classification.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.BCEWithLogitsLoss()

    def forward(self, inputs):
        logits = self.model(inputs)
        loss = self.xent_loss(logits, inputs['y'])
        return {'loss_value':loss}


class MSELoss(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.loss = nn.MSELoss()

    def forward(self, inputs):
        logits = self.model(inputs)
        # transform the inputs['y'] to the float type
        y = inputs['y'].float()
        loss = self.loss(logits, y)
        return {'loss_value':loss}


'''
Implement Logistic Regression model for clinical trial outcome prediction
'''
import os
import pdb

import numpy as np 
from sklearn.linear_model import LogisticRegression as lr_model
import joblib
from rdkit import Chem 
from rdkit.Chem import AllChem
from rdkit.Chem import DataStructs

from pytrial.data.trial_data import TrialOutcomeDatasetBase
from .base import TrialOutcomeBase
from .model_utils.utils import trial_collate_fn

class LogisticRegression(TrialOutcomeBase):
    '''
    Implement Logistic Regression model for clinical trial outcome prediction. 

    Parameters
    ----------
    C: float
        Regularization strength for l2 norm; must be a positive float. 
        Like in support vector machines, smaller values specify weaker regularization.

    dual: bool
        Dual or primal formulation. Dual formulation is only implemented for l2 penalty with liblinear solver. 
        Prefer `dual=False` when `n_samples > n_features`.
    
    solver: {'newton-cg','lbfgs','liblinear','sag','saga'}
        Algorithm to use in the optimization problem. default='lbfgs'.

    max_iter: int (default=100)
        Maximum number of iterations taken for the solvers to converge.
    '''
    def __init__(
        self, 
        C=1.0, 
        dual=False,
        solver='lbfgs', 
        max_iter=100
        ):

        super(LogisticRegression, self).__init__()
        self.model = lr_model(
            C = C,
            dual = dual,
            solver = solver,
            max_iter = max_iter,
        )

    def fit(self, train_data, valid_data):
        '''
        Train logistic regression model to predict clinical trial outcomes.

        Parameters
        ----------
        train_data: TrialOutcomeDatasetBase
            Training data, should be a `TrialOutcomeDatasetBase` object.
        
        valid_data: TrialOutcomeDatasetBase
            Validation data, should be a `TrialOutcomeDatasetBase` object. Ignored for logistic regression model.
            Keep this parameter for compatibility with other models.

        '''
        # build dataloader using train_data
        trainloader = self._build_dataloader_from_dataset(train_data, 
            num_workers=0, batch_size=32, shuffle=True,
            collate_fn=trial_collate_fn)

        _, features, labels = self.feature(trainloader)
        
        self.model.fit(features, labels)

    def predict(self, test_data):
        '''
        Make clinical trial outcome predictions.
        
        Parameters
        ----------
        test_data: TrialOutcomeDatasetBase
            Testing data, should be a `TrialOutcomeDatasetBase` object.
        '''
        # build dataloader using test_data
        testloader = self._build_dataloader_from_dataset(test_data, 
            num_workers=0, batch_size=32, shuffle=False,
            collate_fn=trial_collate_fn)

        nctids, features, labels = self.feature(testloader)

        ypred = self.model.predict_proba(features)
        ypred = ypred[:, 1].tolist() 
        prediction = list(zip(nctids, ypred))
        return prediction

    def save_model(self, output_dir = None):
        '''
        Save the learned logistic regression model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The output folder to save the learned model.
            If set None, will save model to `save_model/model.ckpt`.
        '''

        if output_dir is None:
            output_dir = 'save_model'
        
        if not os.path.exists(output_dir): os.makedirs(output_dir)
        
        filename = os.path.join(output_dir, 'model.ckpt')
        joblib.dump(self.model, filename)

    def load_model(self, checkpoint=None):
        '''
        Load the learned model from the disk.

        Parameters
        ----------
        checkpoint: str 
            The checkpoint folder to load the learned model.
            The checkpoint under this folder should be `model.ckpt`.
        '''
        if checkpoint is None:
            ckpt_dir = 'save_model'
            checkpoint = os.path.join(ckpt_dir, 'model.ckpt')
        else:
            checkpoint = os.path.join(checkpoint, 'model.ckpt')
        self.model = joblib.load(checkpoint)

    def feature(self, data_loader):
        labels = []
        features = []
        nctids = []
        for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in data_loader:
            
            labels.extend(label_vec.tolist())
            nctids.extend(nctid_lst)
            for smiles_lst in smiles_lst2: 
                mol_lst = [Chem.MolFromSmiles(smiles) for smiles in smiles_lst]
                mol_lst = list(filter(lambda x:x is not None, mol_lst))
                fp_lst = [AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=512) for mol in mol_lst]
                feature = np.zeros(512)
                for fp in fp_lst:
                    arr = np.zeros((0,), dtype=np.int8)
                    DataStructs.ConvertToNumpyArray(fp,arr)
                    feature += arr 
                feature = feature.reshape(1,-1)
                features.append(feature)
        features = np.concatenate(features, 0)
        labels = np.array(labels)	
        return nctids, features, labels 






import warnings
from collections import defaultdict
import pdb
import os
from typing import Any, Callable, Dict, List, NewType, Optional, Tuple, Union

import pandas as pd
import torch
from torch.utils.data import Dataset, DataLoader

from ..utils.trial_utils import ClinicalTrials
from ..utils.tabular_utils import read_csv_to_df
from ..data.vocab_data import Vocab

from ..utils.tabular_utils import HyperTransformer

class TrialDatasetBase(Dataset):
    '''
    The basic trial datasets loader.

    Parameters
    ----------
    data: pd.DataFrame
        Contain the trial document in tabular format.

    criteria_column: str
        The column name of eligibility criteria in the dataframe.
    '''
    inc_ec_embedding = None # inclusion criteria embedding
    inc_vocab = None # inclusion criteria vocab
    exc_ec_embedding = None # exclusion criteria embedding
    exc_vocab = None # exclusion criteria vocab

    def __init__(self, data, criteria_column='criteria'):
        self.df = data
        self._process_ec(criteria_column=criteria_column)
        self._collect_cleaned_sentence_set()
    
    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        return self.df.iloc[index:index+1]
    
    def get_ec_sentence_embedding(self):
        '''
        Process the eligibility criteria of each trial,
        get the criterion-level emebddings stored in dict.

        Parameters
        ----------
        criteria_column: str
            The column name of eligibility criteria in the dataframe.
        '''
        if self.inc_ec_embedding is None or self.exc_ec_embedding is None:
            self._get_ec_emb()
        
        return self.inc_ec_embedding, self.exc_ec_embedding

    def _process_ec(self, criteria_column):
        res = self.df[criteria_column].apply(lambda x: self._split_protocol(x))
        self.df['inclusion_criteria'] = res.apply(lambda x: x[0])
        self.df['exclusion_criteria'] = res.apply(lambda x: x[1])

    def _get_ec_emb(self):
        # create EC embedding with indexed ECs
        from pytrial.model_utils.bert import BERT
        # check if cuda is available
        if torch.cuda.is_available():
            device = 'cuda:0'
        else:
            device = 'cpu'
        bert_model = BERT(device=device)
        self.inc_ec_embedding = bert_model.encode(self.inc_vocab.words, batch_size=64)
        self.exc_ec_embedding = bert_model.encode(self.exc_vocab.words, batch_size=64)
        self.inc_ec_embedding = self.inc_ec_embedding.cpu()
        self.exc_ec_embedding = self.exc_ec_embedding.cpu()

    def _collect_cleaned_sentence_set(self):
        # create a vocab for ec sentences
        self.inc_vocab = Vocab()
        self.exc_vocab = Vocab()
        self.inc_vocab.add_sentence(['[PAD]']) # 0 belongs to the pad token
        self.exc_vocab.add_sentence(['[PAD]']) # 0 belongs to the pad token

        inc_index_set, exc_index_set = [], []
        for idx, row in self.df.iterrows():
            row_inc_set, row_exc_set = [], []
            inc = row['inclusion_criteria']
            exc = row['exclusion_criteria']
            for sent in inc:
                self.inc_vocab.add_sentence(sent)
                row_inc_set.append(self.inc_vocab.word2idx[sent])
            for sent in exc:
                self.exc_vocab.add_sentence(sent)
                row_exc_set.append(self.exc_vocab.word2idx[sent])
            inc_index_set.append(list(set(row_inc_set)))
            exc_index_set.append(list(set(row_exc_set)))
        self.df['inclusion_criteria_index'] = inc_index_set
        self.df['exclusion_criteria_index'] = exc_index_set

    def _clean_protocol(self, protocol):
        protocol = protocol.lower()
        protocol_split = protocol.split('\n')
        filter_out_empty_fn = lambda x: len(x.strip())>0
        strip_fn = lambda x: x.strip()
        protocol_split = list(filter(filter_out_empty_fn, protocol_split))	
        protocol_split = list(map(strip_fn, protocol_split))
        return protocol_split

    def _split_protocol(self, protocol):
        protocol_split = self._clean_protocol(protocol)
        inclusion_idx, exclusion_idx = len(protocol_split), len(protocol_split)	
        for idx, sentence in enumerate(protocol_split):
            if "inclusion" in sentence:
                inclusion_idx = idx
                break
        for idx, sentence in enumerate(protocol_split):
            if "exclusion" in sentence:
                exclusion_idx = idx 
                break 		
        if inclusion_idx + 1 < exclusion_idx + 1 < len(protocol_split):
            inclusion_criteria = protocol_split[inclusion_idx:exclusion_idx]
            exclusion_criteria = protocol_split[exclusion_idx:]
            if not (len(inclusion_criteria) > 0 and len(exclusion_criteria) > 0):
                print(len(inclusion_criteria), len(exclusion_criteria), len(protocol_split))
                exit()
            return inclusion_criteria, exclusion_criteria ## list, list 
        else:
            return protocol_split, []


class TrialDataset(Dataset):
    '''
    Basic trial datasets loader.

    Parameters
    ----------
    input_dir: str
        The path to the trial dataset in tabular form (.csv).
        If a directory is given, the code will automatically pick the only '.csv' file under this dir.
    '''
    def __init__(self, input_dir=None) -> None:
        if os.path.isfile(input_dir):
            self.df = read_csv_to_df(input_dir, index_col=0)

        if os.path.isdir(input_dir):
            csv_names = [name for name in os.listdir(input_dir) if name.endswith('.csv')]
            if len(csv_names) > 1:
                raise Exception(f'`input_dir` {input_dir} is given where more than one csv files are found under this path.')
            if len(csv_names) == 0:
                raise Exception(f'`input_dir` {input_dir} is given where no csv file is found under this path.')
            self.df = read_csv_to_df(os.path.join(input_dir, csv_names[0]), index_col=0)

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        return self.df.iloc[index:index+1]

class TrialDataCollator:
    '''The basic trial data collator.
    Subclass it and override the `__init__` & `__call__` function if need operations inside this step.

    Returns
    -------
    batch_df: pd.DataFrame
        A dataframe contains multiple fields for each trial.
    '''
    def __init__(self) -> None:
        # subclass to add tokenizer
        # subclass to add feature preprocessor
        pass

    def __call__(self, examples):
        batch_df = pd.concat(examples, 0)
        batch_df.fillna('none',inplace=True)
        return batch_df


class TrialOutcomeDatasetBase(TrialDatasetBase):
    '''
    Basic trial outcome datasets loader.

    Parameters
    ----------
    data: pd.DataFrame
        Contain the trial document in tabular format.
    '''
    columns = ['nctid', 'label', 'smiless',  'icdcodes', 'criteria']
    def __init__(self, data, columns=None) -> None:
        self.data = data
        if columns is not None:
            self.columns = columns
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        row = self.data.iloc[index]
        return row[self.columns[0]], row[self.columns[1]], row[self.columns[2]], row[self.columns[3]], row[self.columns[4]]


class TrialDatasetStructured(Dataset):
    '''
    Dataset class for structured trial features. Subclass it if additional properties and functions
    are required to add for specific tasks. We make use `rdt`: https://docs.sdv.dev/rdt for transform
    and reverse transform of the tabular data.

    Parameters
    ----------
    df: pd.DataFrame
        The input trial tabular format records.

    metadata: dict
        Contains the meta setups of the input data. It should contain the following keys:

        (1) `sdtypes`: dict, the data types of each column in the input data. The keys are the column
            names and the values are the data types. The data types can be one of the following:
            'numerical', 'categorical', 'datetime', 'boolean'.

        (2) `transformers`: dict, the transformers to be used for each column. The keys are the column
            names and the values are the transformer names. The transformer names can be one in
            https://docs.sdv.dev/rdt/transformers-glossary/browse-transformers.

        metadata = {

        'sdtypes': {

        'column1': 'numerical',
        'column2': 'boolean',
        'column3': 'datetime',
        'column4': 'categorical',
        'column5': 'categorical',

        },

        transformers':{

        'column1': rdt.transformers.FloatFormatter(missing_value_replacement='mean'),
        'column2': rdt.transformers.BinaryEncoder(missing_value_replacement='mode'),
        'column3': rdt.transformers.UnixTimestampEncoder(missing_value_replacement='mean'),
        'column4': rdt.transformers.FrequencyEncoder(),
        'column5': None, # will not do any thing to this column if no transformer is specified.
        
        }
        
        } 
        
        It is recommend to provide the metadata of the input tabular data.

        - If no metdata is given, the dataset will automatically detect the `dtypes` of columns and build the corresponding `transformers`.
        - If only `sdtypes` are given, the dataset will detect if there are missing `sdtypes` given and build the `transformers` and dtype automatically.

    transform: bool(default=True)
        Whether or not transform raw self.df by hypertransformer.
        If set False, :code:`self.df` will keep as the same as the passed one.

    Examples
    --------
    >>> from pytrial.data.site_data import TabularSiteBase
    >>> df = pd.read_csv('tabular_trial.csv', index_col=0)
    >>> # set `transform=True` will replace dataset.df with dataset.df_transformed
    >>> dataset = TabularPatientBase(df, transform=True)
    >>> # transform raw dataframe to numerical tables
    >>> df_transformed = dataset.transform(df)
    >>> # make back transform to the original df
    >>> df_raw = dataset.reverse_transform(df_transformed)
    '''
    def __init__(self, df, metadata=None, transform=True):
        self.df = df
        self.metadata = metadata

        # initialize hypertransformer
        self.ht = HyperTransformer()

        if transform:
            if metadata is None:
                warnings.warn('No metadata provided. Metadata will be automatically '
                            'detected from your data. This process may not be accurate. '
                            'We recommend writing metadata to ensure correct data handling.')
                self.ht.detect_initial_config(df)
                self.metadata = self.ht.get_config()
                self.ht.fit(df)

            else:
                # parse the metadata and update hypertransformer's config
                self._parse_metadata()

            # replace data with the transformed one
            self.df = self.transform(df)

    def __getitem__(self, index):
        # TODO: support better indexing
        '''
        Indexing the dataframe stored in tabular patient dataset.

        Parameters
        ----------
        index: int or list[int]
            Retrieve the corresponding rows in the dataset.
        '''
        if isinstance(index, int):
            return self.df.iloc[index:index+1]
        
        elif isinstance(index, list):
            return self.df.iloc[index]

    def __len__(self):
        return len(self.df)

    def __repr__(self):
        return f'<pytrial.data.trial_data.TrialDatasetStructured object> Tabular trial data with {self.df.shape[0]} samples {self.df.shape[1]} features, call `.df` to yield the pd.DataFrame data: \n' + repr(self.df)

    def transform(self, df=None):
        '''
        Transform the input df or the self.df by hypertransformer.
        If transform=True in `__init__`, then you do not need to call this function
        to transform self.df because it was tranformed already.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed by self.ht
        '''

        if df is None:
            return self.ht.transform(self.df)
        else:
            return self.ht.transform(df)

    def reverse_transform(self, df=None):
        '''
        Reverse the input dataframe back to the original format. Return the self.df in the original
        format if `df=None`.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed back to the original format by self.ht.
        '''
        if df is None:
            return self.ht.reverse_transform(self.df)
        else:
            return self.ht.reverse_transform(df)

    def _parse_metadata(self):
        '''
        Parse the passed metadata, cope with the following scnearios:
        (1) only `sdtypes` are given;
        (2) only `transformers` are given;
        (3) only partial `sdtypes` are given;
        (4) only partial `transformers` are given.
        '''
        # parse metadata dict for building the hypertransformer
        metadata = self.metadata
        self.ht.detect_initial_config(self.df, verbose=False)

        if 'transformers' in metadata:
            self.ht.update_transformers(metadata['transformers'])
        
        if 'sdtypes' in metadata:
            self.ht.update_sdtypes(metadata['sdtypes'])

        self.ht.fit(self.df)
        self.metadata.update(self.ht.get_config())

def test():
    trialdata = TrialDataset('./datasets/AACT-ClinicalTrial/')
    trial_collate_fn = TrialDataCollator()
    trialoader = DataLoader(trialdata, batch_size=10, shuffle=False, collate_fn=trial_collate_fn)
    batch = next(iter(trialoader))
    print(batch)

if __name__ == '__main__':
    test()



import pdb

import numpy as np
import torch
from torch import nn
from torch.autograd import grad as torch_grad

class XentLoss(nn.Module):
    '''
    The basic xent loss
    for multi-class classification.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.CrossEntropyLoss()
    
    def forward(self, inputs):
        logits = self.model(inputs['x'])
        loss = self.xent_loss(logits, inputs['y'])
        return {'loss_value':loss}

class BinaryXentLoss(nn.Module):
    '''
    The basic binary xent loss
    for binary classification.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.BCEWithLogitsLoss()

    def forward(self, inputs):
        logits = self.model(inputs['x'])
        y = inputs['y']
        if y.shape != logits.shape:
            logits = logits.reshape(y.shape)
        loss = self.xent_loss(logits, inputs['y'])
        return {'loss_value':loss}

class MultilabelBinaryXentLoss(nn.Module):
    '''
    The basic binary xent loss
    for multilabel classification.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.BCEWithLogitsLoss()

    def forward(self, inputs):
        logits, labels = self.model(inputs)
        loss = self.xent_loss(logits, labels)
        return {'loss_value':loss}
    
class MultilabelBinaryXentLossWithKLDivergence(nn.Module):
    '''
    The basic binary xent loss
    for multilabel classification with
    KL Divergence for VAE regularization.
    '''
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.xent_loss = nn.BCEWithLogitsLoss()

    def forward(self, inputs, kl_weight=1.0):
        logits, labels, kl_loss = self.model(inputs)
        loss = self.xent_loss(logits, labels)
        tot_loss = loss + (kl_weight * kl_loss)
        return {'loss_value': tot_loss}

class MSELoss(nn.Module):
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
        self.loss = nn.MSELoss()

    def forward(self, inputs):
        logits = self.model(inputs['x'])
        loss = self.loss(logits, inputs['y'])
        return {'loss_value':loss}


'''
Two GAN based loss for GAN-based generator.
'''
class GeneratorLoss(nn.Module):
    name = 'generator_loss'
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model

    def forward(self, inputs):
        outputs = self.model(inputs)
        y_fake = outputs['y_fake']

        # naive GAN loss
        # loss_g = - y_fake.log().mean()

        # WGAN loss, more stable
        loss_g = -torch.mean(y_fake)

        return {'loss_value':loss_g}

class DiscriminatorLoss(nn.Module):
    name = 'discriminator_loss'
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
    
    def forward(self, inputs):
        outputs = self.model(inputs)
        y_real = outputs['y_real']
        y_fake = outputs['y_fake']

        # naive GAN loss
        # loss_d = - (1 - y_fake).log().mean() - y_real.log().mean()

        # WGAN loss, more stable
        loss_d = torch.mean(y_fake) - torch.mean(y_real)

        return {'loss_value':loss_d}

class DiscriminatorLossGP(nn.Module):
    name = 'discriminator_loss'
    def __init__(self, model) -> None:
        super().__init__()
        self.model = model
    
    def forward(self, inputs):
        outputs = self.model(inputs)
        y_real = outputs['y_real']
        y_fake = outputs['y_fake']

        # naive GAN loss
        # loss_d = - (1 - y_fake).log().mean() - y_real.log().mean()

        # WGAN loss, more stable
        loss_d = torch.mean(y_fake) - torch.mean(y_real)

        # gradient penalty
        gp = self.compute_gradient_penalty(outputs)
        loss_d += gp

        return {'loss_value':loss_d}

    def compute_gradient_penalty(self, outputs):
        '''
        Requires `self.model.infer_discriminator` implemented.
        '''
        if 'interpolates' in outputs:
            interpolates = outputs['interpolates']
            d_interpolates = outputs['d_interpolates']
        else:
            x_fake = outputs['x_fake']
            x_real = outputs['y']
            alpha = torch.tensor(np.random.randn(x_real.size(0))).to(x_fake.device)
            interpolates = (alpha * x_real + (1-alpha)*x_fake).requires_grad_(True).float()
            d_interpolates = self.model.infer_discriminator(interpolates)
        grad_outputs=torch.ones(d_interpolates.size()).to(d_interpolates.device)
        gradients = torch_grad(outputs=d_interpolates, inputs=interpolates, grad_outputs=grad_outputs, create_graph=True, retain_graph=True)[0]
        gradients = gradients.view(gradients.size(0), -1)
        gradient_penalty = ((gradients.norm(2, dim=1) - 1) ** 2).mean()
        return gradient_penalty



import pdb

import torch
from torch import nn

from pytrial.utils.check import (
    check_checkpoint_file, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceIndivBase
from ..trainer import IndivSeqTrainer


class BuildModel(nn.Module):
    def __init__(self, 
        input_dim = None, 
        hidden_dim = 384, 
        conv_size = 10, 
        levels = 3, 
        dropconnect = 0.3, 
        dropout = 0.3, 
        dropres = 0.3,
        label_size = None,
        device = None
        ):
        super(BuildModel, self).__init__()
        assert hidden_dim % levels == 0
        self.dropout = dropout
        self.dropconnect = dropconnect
        self.dropres = dropres
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.conv_dim = hidden_dim
        self.conv_size = conv_size
        self.output_dim = label_size
        self.levels = levels
        self.chunk_size = hidden_dim // levels
        self.device = device

        self.kernel = nn.Linear(int(input_dim+1), int(hidden_dim*4+levels*2))
        nn.init.xavier_uniform_(self.kernel.weight)
        nn.init.zeros_(self.kernel.bias)
        self.recurrent_kernel = nn.Linear(int(hidden_dim+1), int(hidden_dim*4+levels*2))
        nn.init.orthogonal_(self.recurrent_kernel.weight)
        nn.init.zeros_(self.recurrent_kernel.bias)

        self.nn_scale = nn.Linear(int(hidden_dim), int(hidden_dim // 6))
        self.nn_rescale = nn.Linear(int(hidden_dim // 6), int(hidden_dim))
        self.nn_conv = nn.Conv1d(int(hidden_dim), int(self.conv_dim), int(conv_size), 1)
        self.nn_output = nn.Linear(int(self.conv_dim), int(self.output_dim))

        if self.dropconnect:
            self.nn_dropconnect = nn.Dropout(p=dropconnect)
            self.nn_dropconnect_r = nn.Dropout(p=dropconnect)
        if self.dropout:
            self.nn_dropout = nn.Dropout(p=dropout)
            self.nn_dropres = nn.Dropout(p=dropres)

    def cumax(self, x, mode='l2r'):

        if mode == 'l2r':
            x = torch.softmax(x, dim=-1)
            x = torch.cumsum(x, dim=-1)
            return x
        elif mode == 'r2l':
            x = torch.flip(x, [-1])
            x = torch.softmax(x, dim=-1)
            x = torch.cumsum(x, dim=-1)
            return torch.flip(x, [-1])
        else:
            return x

    def step(self, inputs, c_last, h_last, interval):

        x_in = inputs
        # Integrate inter-visit time intervals
        interval = interval.unsqueeze(-1)
        x_out1 = self.kernel(torch.cat((x_in, interval), dim=-1))
        x_out2 = self.recurrent_kernel(torch.cat((h_last, interval), dim=-1))

        if self.dropconnect:
            x_out1 = self.nn_dropconnect(x_out1)
            x_out2 = self.nn_dropconnect_r(x_out2)
        x_out = x_out1 + x_out2
        f_master_gate = self.cumax(x_out[:, :self.levels], 'l2r')
        f_master_gate = f_master_gate.unsqueeze(2)
        i_master_gate = self.cumax(x_out[:, self.levels:self.levels*2], 'r2l')
        i_master_gate = i_master_gate.unsqueeze(2)
        x_out = x_out[:, self.levels*2:]
        x_out = x_out.reshape(-1, self.levels*4, self.chunk_size)
        f_gate = torch.sigmoid(x_out[:, :self.levels])
        i_gate = torch.sigmoid(x_out[:, self.levels:self.levels*2])
        o_gate = torch.sigmoid(x_out[:, self.levels*2:self.levels*3])
        c_in = torch.tanh(x_out[:, self.levels*3:])
        c_last = c_last.reshape(-1, self.levels, self.chunk_size)
        overlap = f_master_gate * i_master_gate
        c_out = overlap * (f_gate * c_last + i_gate * c_in) + (f_master_gate - overlap) * c_last + (i_master_gate - overlap) * c_in
        h_out = o_gate * torch.tanh(c_out)
        c_out = c_out.reshape(-1, self.hidden_dim)
        h_out = h_out.reshape(-1, self.hidden_dim)
        out = torch.cat([h_out, f_master_gate[..., 0], i_master_gate[..., 0]], 1)
        return out, c_out, h_out

    def forward(self, input_data):
        
        """
        
        Parameters
        
        ----------
        input_data = {
                      'X': shape (batchsize, n_timestep, n_featdim)
                      'M': shape (batchsize, n_timestep)
                      'cur_M': shape (batchsize, n_timestep)
                      'T': shape (batchsize, n_timestep), the time interval between visits, i.e., for timestamp [0, 1, 2, 3, 4], T = [0, 1, 1, 1, 1].
                     }
        
        Return
        
        ----------
        
        all_output, shape (batchsize, n_timestep, n_labels)
            
            predict output of each time step
            
        cur_output, shape (batchsize, n_labels)
        
            predict output of last time step
        
        """
        X = input_data['v']

        if 'timestamp' not in input_data:
            # if no timestamp is provided, we assume the time interval between visits is 1.
            T = torch.ones_like(X[:, :, 0])
            T[:,0] = 0
            T = T.to(X.device)
        else:
            T = input_data['timestamp']
            # differentiate between timestamp
            T = T[:,1:] - T[:,:-1]
            T[:,0] = 0

        if 'v_lengths' in input_data:
            v_len = input_data['v_lengths']
            mask, cur_mask = self._create_mask_from_visit_length(v_len=v_len)
            mask = mask.to(X.device)
            cur_mask = cur_mask.to(X.device)
        else:
            mask = cur_mask = None

        batch_size, time_step, feature_dim = X.size()
        c_out = torch.zeros(batch_size, self.hidden_dim).to(self.device)
        h_out = torch.zeros(batch_size, self.hidden_dim).to(self.device)
        tmp_h = torch.zeros_like(h_out, dtype=torch.float32).view(-1).repeat(self.conv_size).view(self.conv_size, batch_size, self.hidden_dim).to(self.device)
        tmp_dis = torch.zeros((self.conv_size, batch_size)).to(self.device)
        h = []
        origin_h = []
        distance = []

        for t in range(time_step):
            out, c_out, h_out = self.step(X[:, t, :], c_out, h_out, T[:, t])
            cur_distance = 1 - torch.mean(out[..., self.hidden_dim:self.hidden_dim+self.levels], -1)
            cur_distance_in = torch.mean(out[..., self.hidden_dim+self.levels:], -1)
            origin_h.append(out[..., :self.hidden_dim])
            tmp_h = torch.cat((tmp_h[1:], out[..., :self.hidden_dim].unsqueeze(0)), 0)
            tmp_dis = torch.cat((tmp_dis[1:], cur_distance.unsqueeze(0)), 0)
            distance.append(cur_distance)
            #Re-weighted convolution operation
            local_dis = tmp_dis.permute(1, 0)
            local_dis = torch.cumsum(local_dis, dim=1)
            local_dis = torch.softmax(local_dis, dim=1)
            local_h = tmp_h.permute(1, 2, 0)
            local_h = local_h * local_dis.unsqueeze(1)
            #Re-calibrate Progression patterns
            local_theme = torch.mean(local_h, dim=-1)
            local_theme = self.nn_scale(local_theme)
            local_theme = torch.relu(local_theme)
            local_theme = self.nn_rescale(local_theme)
            local_theme = torch.sigmoid(local_theme)
            local_h = self.nn_conv(local_h).squeeze(-1)
            local_h = local_theme * local_h
            h.append(local_h)  

        origin_h = torch.stack(origin_h).permute(1, 0, 2)
        rnn_outputs = torch.stack(h).permute(1, 0, 2)
        if self.dropres > 0.0:
            origin_h = self.nn_dropres(origin_h)
        rnn_outputs = rnn_outputs + origin_h
        rnn_outputs = rnn_outputs.contiguous().view(-1, rnn_outputs.size(-1))
        if self.dropout > 0.0:
            rnn_outputs = self.nn_dropout(rnn_outputs)
        output = self.nn_output(rnn_outputs)
        output = output.contiguous().view(batch_size, time_step, self.output_dim)

        cur_output = (output * cur_mask.unsqueeze(-1)).sum(dim=1)
        all_output = torch.sigmoid(output)

        # TODO: all_output is the per visit label prediction, currently we only support patient-level prediction.
        # return all_output, cur_output
        return cur_output

    def _create_mask_from_visit_length(self, v_len):
        mask = torch.zeros(len(v_len),max(v_len))
        cur_mask = torch.zeros(len(v_len),max(v_len))

        for i in range(len(v_len)):
            mask[i, :v_len[i]] = 1
            cur_mask[i, v_len[i]-1] = 1
        return mask, cur_mask

class StageNet(SequenceIndivBase):
    '''
    Implement StageNet for longitudinal patient records predictive modeling [1]_.

    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    orders: list[str]
        A list of orders when treating inputs events. Should have the same shape of `vocab_size`.

    mode: str
        Prediction traget in ['binary','multiclass','multilabel','regression'].

    output_dim: int
        If binary classification, output_dim=1;
        If multiclass/multilabel classification, output_dim=n_class
        If regression, output_dim=1.

    max_visit: int
        The maximum number of visits for input event codes.

    hidden_size : int, optional (default = 8)
        The number of features of the hidden state h.
    
    conv_size : int, optional (default = 10)
        The number of convolution kernels.
    
    levels : int, optional (default = 3)
        The number of levels for the master gate.
    
    dropconnect : float, optional (default = 0.3)
        The dropout rate for the input of the convolutional layer.

    dropout : float, optional (default = 0.3)
        The dropout rate for the output of the RNN layer.
    
    dropres : float, optional (default = 0.3)
        The dropout rate for the residual connection.
    
    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weight_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.

    
    Notes
    -----
    .. [1] Gao, J., Xiao, C., Wang, Y., Tang, W., Glass, L. M., & Sun, J. (2020, April). Stagenet: Stage-aware neural networks for health risk prediction. 
        In Proceedings of The Web Conference 2020 (pp. 530-540).
    '''
    def __init__(self,
        vocab_size,
        orders,
        mode,
        output_dim=None,
        max_visit=None,
        hidden_size=384,
        conv_size=10,
        levels=3,
        dropconnect=0.3,
        dropout=0.3,
        dropres=0.3,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ):
        super().__init__(experiment_id, mode=mode, output_dim=output_dim)
        self.config = {
            'mode':self.mode,
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'hidden_size':hidden_size,
            'output_dim':self.output_dim,
            'dropout':dropout,
            'dropres':dropres,
            'dropconnect':dropconnect,
            'conv_size':conv_size,
            'levels':levels,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':orders,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()

    def predict(self, test_data):
        '''
        Predict patient outcomes using longitudinal trial patient sequences.
        
        Parameters
        ----------
        test_data: SequencePatient
            A `SequencePatient` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        '''
        test_dataloader = self.get_test_dataloader(test_data)
        outputs = self._predict_on_dataloader(test_dataloader)
        return outputs['pred']

    def fit(self, train_data, valid_data):
        '''
        Train model with sequential patient records.
        
        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events; 'y' corresponds to labels.

        valid_data: SequencePatientBase
            A `SequencePatientBase` contains patient records used to make early stopping of the
            model.
        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data=valid_data)

    def load_model(self, checkpoint):
        '''
        Load pretrained model from the disk.

        Parameters
        ----------
        checkpoint: str
            The input directory that stores the trained pytorch model and configuration.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        
        self.model.load_state_dict(state_dict['model'])

    def save_model(self, output_dir):
        '''
        Save the pretrained model to the disk.

        Parameters
        ----------
        output_dir: str
            The output directory that stores the trained pytorch model and configuration.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model.state_dict()}, output_dir=output_dir)

    def _build_model(self):
        self.model = BuildModel(
            input_dim=self.config['total_vocab_size'],
            hidden_dim=self.config['hidden_size'],
            conv_size=self.config['conv_size'],
            levels=self.config['levels'],
            dropout=self.config['dropout'],
            dropres=self.config['dropres'],
            dropconnect=self.config['dropconnect'],
            label_size=self.config['output_dim'],
            device=self.config['device'],
            )
        self.model.to(self.device)

    def _fit_model(self, train_data, valid_data=None):
        test_metric_dict = {
            'binary': 'auc',
            'multiclass': 'acc',
            'regression': 'mse',
            'multilabel': 'f1', # take average of F1 scores
            }
        train_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model) for loss_model in loss_models]

        if self.config['mode'] == 'regression':
            less_is_better = True
        else:
            less_is_better = False

        trainer = IndivSeqTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,
            test_metric=test_metric_dict[self.config['mode']],
            less_is_better=less_is_better,
            )
        trainer.train(**self.config)


import pdb
import os
import pickle

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from torch import nn
from datasets import Dataset
from tqdm import tqdm

from pytrial.model_utils.bert import BERT
from pytrial.utils.check import make_dir_if_not_exist
from .base import TrialSearchBase
from .base import whitening_torch_final
from ..metrics import precision, recall, ndcg
from ..data import TrialSearchCollator, batch_to_device


class BuildModel:
    def __new__(self, bert_name):
        return BERT(bert_name)    

class WhitenBERT(TrialSearchBase):
    '''
    Implement a postprocessing method to improve BERT embeddings for similarity search [1]_.

    Parameters
    ----------
    layer_mode: {'last_first', 'last''}
        The mode of layer of embeddings to aggregate. 'last_first' means use the last layer and the first layer. 'last' means use the last layer only.

    bert_name: str, optional (default = 'emilyalsentzer/Bio_ClinicalBERT')
        The name of base BERT model used for encoding input texts.

    device: str, optional (default = 'cuda:0')
        The device of this model, typically be 'cpu' or 'cuda:0'.

    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    
    Notes
    -----
    .. [1] Huang, J., Tang, D., Zhong, W., Lu, S., Shou, L., Gong, M., ... & Duan, N. (2021, November). WhiteningBERT: An Easy Unsupervised Sentence Embedding Approach. In Findings of the Association for Computational Linguistics: EMNLP 2021 (pp. 238-244).
    '''
    trial_embs = {}
    def __init__(self,
        layer_mode='last_first',
        bert_name='emilyalsentzer/Bio_ClinicalBERT',
        device='cuda:0',
        experiment_id='test'):
        super().__init__(experiment_id)
        self.layer_mode = layer_mode
        self.bert_name = bert_name
        self.device = device
        self._build_model()
        self.config = {
            'layer_model': self.layer_mode,
            'bert_name': self.bert_name,
            'device': self.device,
            'batch_size': 32, # default batch size for encoding
            'num_workers': 4, # default num_workers for encoding
            'max_seq_length': 512, # default max_seq_length for encoding
            'tag_field': 'nctid', # default tag_field for encoding
        }

    def __getitem__(self, tag):
        return self.embs[tag]

    def fit(self, train_data, valid_data=None):
        '''
        Go over all trials and encode them into embeddings.
        Note that this is a post-processing method based on a pretrained BERT model,
        so it does `NOT` need to be trained.

        Parameters
        ----------
        train_data: dict
            The data for encoding.

            - 'x' is the dataframe that contains multiple sections of a trial.
            - 'fields' is the list of fields to be encoded.
            - 'tag' is the unique index column name of each document, e.g., 'nctid'.

            train_data = 
            
            {
            
            'x': pd.DataFrame,

            'fields': list[str],

            'tag': str,

            }

        valid_data: Not used.
            This is a placeholder because this model does not need training.
        '''
        embs = self.encode(train_data, return_dict=True)
        self.trial_embs = embs
        return embs

    def predict(self,
        test_data,
        top_k=10,
        return_df=True):
        '''
        Predict the top-k relevant for input documents.

        Parameters
        ----------
        test_data: dict
            Share the same input format as the `train_data` in `fit` function.
            If `fields` and `tag` are not given, will reuse the ones used during training.

            test_data = 

            {
            
            'x': pd.DataFrame,

            'fields': list[str],

            'tag': str,
            
            }

        top_k: int
            Number of retrieved candidates.

        return_df: float
            - If set True, return dataframe for the computed similarity ranking.
            - else, return rank_list=[[(doc1,sim1),(doc2,sim2)], [(doc1,sim1),...]].

        Returns
        -------
        rank: pd.DataFrame
            A dataframe contains the top ranked NCT ids for each.

        sim: pd.DataFrame
            A dataframe contains the corresponding similarities.

        rank_list: list[list[tuple]]
            A list of tuples of top ranked docs and similarities.
        '''
        self._input_data_check(test_data)
        
        tag_field = self.config['tag_field']

        assert len(self.trial_embs) > 0, 'No embeddings found. Please run `fit` first. Or load embeddings from `load_model`.'

        all_embs = np.stack(self.trial_embs.values())
        all_tags = np.stack(self.trial_embs.keys())

        # TODO: skip encoding those already stored in model
        # to_encode_test_trial = test_trial[~test_trial[tag_field].isin(all_tags)]
        tags, embs = self.encode(test_data, return_dict=False)
        
        # rank for each test document
        sim = embs.dot(all_embs.T)
        rank = np.argsort(sim, 1)[:,::-1] # flip
        rank = rank[:,1:top_k+1]

        if return_df:
            tag_list, sim_list = [], []
            for i,row in enumerate(rank):
                tag_list.append(all_tags[row])
                sim_list.append(sim[i][row])
            tag_ar = np.array(tag_list)
            sim_ar = np.array(sim_list)
            rank_df = pd.DataFrame(tag_ar, columns=[f'rank_{i}' for i in range(top_k)], index=tags)
            sim_df = pd.DataFrame(sim_ar, columns=[f'rank_{i}' for i in range(top_k)], index=tags)
            return rank_df, sim_df
        else:
            rank_list = []
            for i,row in enumerate(rank):
                tag_ar = all_tags[row]
                sim_ar = sim[i][row]
                rank_list.append(
                    list(zip(tag_ar, sim_ar))
                )
            return rank_list

    def evaluate(self, test_data):
        '''
        Evaluate within the given trial and corresponding candidate trials.

        Parameters
        ----------
        test_data:  dict
            The provided labeled dataset for test trials. Follow the format listed below.

            test_data =

            {

            'x': pd.DataFrame,

            'y': pd.DataFrame

            }


        Returns
        -------
        results: dict[float]
            A dict of metrics and the values.

        Notes
        -----
        x =
        
        | target_trial | trial1 | trial2 | trial3 |


        | nct01        | nct02  | nct03  | nct04  |

        y =

        | label1 | label2 | label3 |


        | 0      | 0      | 1      |

        '''
        test_df = test_data['x']
        label_df = test_data['y']
        ranked_label_list = []
        for idx, row in test_df.iterrows():
            target_trial = row['target_trial']
            if target_trial in self.trial_embs:
                target_emb = self.__getitem__(target_trial)
            else:
                raise ValueError(f'The embeddings of trial {target_trial} are not found.')
            candidate_embs = np.stack([self.__getitem__(tag) for tag in row.values[1:]])
            sim = target_emb[None].dot(candidate_embs.T)[0]
            labels = label_df.iloc[idx].to_numpy()
            if labels.sum() == 0: continue
            ranked_label = labels[np.argsort(sim)[::-1]]
            ranked_label_list.append(ranked_label)
        ranked_label_list = np.array(ranked_label_list)

        return_dict = {}
        for k in [1,2,5]:
            return_dict[f'precision@{k}'] = precision(ranked_label_list, k)
            return_dict[f'recall@{k}'] = recall(ranked_label_list, k)
        return_dict[f'ndcg@{k}'] = ndcg(ranked_label_list, k)
        return return_dict

    def save_model(self, output_dir):
        '''
        Only save the embeddings. Do not save the model.
        
        Parameters
        ----------
        output_dir: str
            The output directory to save the model.
        '''
        if not os.path.exists(output_dir):
            make_dir_if_not_exist(output_dir)
        with open(os.path.join(output_dir, 'embeddings.pkl'), 'wb') as f:
            pickle.dump(self.trial_embs, f)
        print(f'Embeddings saved to {output_dir}')

    def load_model(self, input_dir):
        '''
        Only load the embeddings. Do not load the model.
        
        Parameters
        ----------
        input_dir: str
            The input directory to load the model.
        '''
        with open(os.path.join(input_dir, 'embeddings.pkl'), 'rb') as f:
            self.trial_embs = pickle.load(f)
        print(f'Embeddings loaded from {input_dir}')

    def encode(self,
        inputs,
        batch_size=None,
        num_workers=None,
        return_dict=True,
        verbose=True,
        ):
        '''
        Encode input documents and output the document embeddings.

        Parameters
        ----------
        inputs: dict
            The input documents to encode:

            - 'fields' is the list of fields to be encoded.

            - 'tag' is the unique index column name of each document, e.g., 'nctid'.

            inputs = 

            {
            
            'x': pd.DataFrame,

            'fields': list[str],

            'tag': str,

            }

        batch_size: int, optional
            The batch size when encoding trials.
        
        num_workers: int, optional
            The number of workers when building the val dataloader.

        return_dict: bool
            Whether to return a dict of results.

            - If set True, return dict[np.ndarray].
            - Else, return np.ndarray with the order same as the input documents.

        verbose: bool
            Whether plot progress bar or not.

        Returns
        -------
        embs: dict[np.ndarray]
            Encoded trial-level embeddings with key (tag) and value (embedding)..
        '''
        self._input_data_check(inputs)
        if batch_size is not None:
            self.config['batch_size'] = batch_size
        
        if num_workers is not None:
            self.config['num_worker'] = num_workers

        # build dataset and dataloader
        dataloader = self.get_val_dataloader(inputs)

        # encode all trials
        embs = self._encode_for_dataloader(dataloader, verbose=verbose, return_dict=return_dict)
        return embs

    def get_val_dataloader(self, inputs):
        df = inputs['x']
        self.config['fields'] = inputs['fields']
        self.config['tag_field'] = inputs['tag']
        dataset = self._build_val_dataset(df=df)
        collate_fn = self._build_val_collator()
        dataloader = self._build_val_dataloader(dataset=dataset, collate_fn=collate_fn)
        return dataloader

    def _build_model(self):
        self.model = BuildModel(self.bert_name)
        self.model.to(self.device)

    def _encode_for_dataloader(self, dataloader, verbose=True, return_dict=True):
        emb_list = []
        tag_list = []
        self.model.eval()
        with torch.no_grad():
            for data in tqdm(dataloader, desc='Encoding', disable=not verbose):
                if self.config['tag_field'] is not None:
                    tag = data.pop(self.config['tag_field'])
                    tag_list.append(tag)

                data = self._prepare_input(data)

                # get an average of embeddings of all fields
                embs = []
                for field in self.config['fields']:
                    emb = self._encode_for_field(data=data, field=field)
                    embs.append(emb)
                temp_emb = torch.stack(embs, dim=0).mean(dim=0)
                if len(temp_emb.shape) == 1: temp_emb = temp_emb[None]

                # whitening for each batch
                temp_emb = whitening_torch_final(temp_emb)
                emb_list.append(temp_emb.cpu().numpy())

        embs = np.concatenate(emb_list, 0)

        if len(tag_list) == 0:
            tag_list = list(range(len(embs)))
        else:
            tag_list = sum(tag_list, [])

        if return_dict:
            embs = dict(zip(tag_list, embs))
            return embs
        else:
            return (tag_list, embs)

    def _encode_for_field(self, data, field):
        inputs = {}
        for k, v in data.items():
            if field in k:
                tempkey = k.replace(field+'_','')
                inputs[tempkey] = v
        inputs = batch_to_device(inputs, self.device)
        if self.layer_mode == 'last_first':
            # take a mean of the last and first layer
            inputs['return_hidden_states'] = True
            outputs = self.model(**inputs)
            emb = (outputs[1].mean(1) + outputs[-1].mean(1))/2
        else:
            # only take the last layer embedding
            emb = self.model(**inputs)
        return emb

    def _prepare_input(self, inputs):
        fields = self.config['fields']
        output_dict = {}
        for fd in fields:
            if fd not in inputs:
                continue
            for k,v in inputs[fd].items():
                output_dict[f'{fd}_{k}'] = v
        return output_dict

    def _build_val_dataset(self, df):
        dataset = Dataset.from_pandas(df)
        dataset.set_transform(self._val_dataset_transform)
        return dataset

    def _build_val_collator(self):
        return TrialSearchCollator(
            bert_name=self.config['bert_name'],
            max_seq_length=self.config['max_seq_length'],
            fields=self.config['fields'],
            device=self.config['device'],
            tag_field=self.config['tag_field'],
        )

    def _build_val_dataloader(self, dataset, collate_fn):
        dataloader = DataLoader(
            dataset,
            batch_size=self.config['batch_size'],
            shuffle=False,
            num_workers=self.config['num_workers'],
            collate_fn=collate_fn,
            pin_memory=True,
        )
        return dataloader

    def _dataset_transform(self, examples):
        return_dict = {}

        if self.config['fields'] is not None:
            for k in self.config['fields']: 
                return_dict[k] = examples[k]
        
        if self.config['tag_field'] is not None:
            return_dict[self.config['tag_field'] ] = examples[self.config['tag_field'] ]

        if len(return_dict) == 0:
            return_dict = examples

        return return_dict

    def _val_dataset_transform(self, examples):
        return self._dataset_transform(examples=examples)

    def _input_data_check(self, inputs):
        '''
        Check the training / testing data fits the formats.
        Target to (1) check if inputs valid,
                    if not, give tips about the data problem.

        Parameters
        ----------
        inputs: {
                'x': pd.DataFrame,
                'fields': list[str],
                'tag': str,
                }
        '''
        # check overall input format
        assert 'x' in inputs, 'No input trial doc dataframe found in inputs.'
        df = inputs['x']
        if 'fields' in inputs:
            try:
                _ = df[inputs['fields']]
            except:
                raise Exception('Cannot find the specified `fields` in inputs dataframe.')
        if 'tag' in inputs:
            try:
                _ = df[inputs['tag']]
            except:
                raise Exception('Cannot find the specified `tag` in inputs dataframe.')

        # check data type
        try:
            _ = df.applymap(str)
        except:
            raise Exception('Cannot transform the input dataframe to str type, please check the inputs.')


'''
Implement dataset for trial search model training.
'''
import pdb
from collections import defaultdict

import pandas as pd
from transformers import AutoTokenizer
from torch import Tensor, device

from pytrial.data.trial_data import TrialDataCollator

class TrialSearchCollator(TrialDataCollator):
    '''
    The basic collator for trial search tasks.
    '''
    def __init__(self,
        bert_name,
        max_seq_length,
        fields,
        device,
        tag_field=None,
        ) -> None:
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(bert_name)
        self.max_length = max_seq_length
        self.fields = fields
        self.device = device
        self.tag_field = tag_field

    def __call__(self, features):
        return_dict = defaultdict(list)
        batch_df = pd.DataFrame(features)
        batch_df.fillna('', inplace=True)

        fields = self.fields
        return_dict.update(self._batch_tokenize(batch_df=batch_df, fields=fields))
        if self.tag_field is not None:
            return_dict[self.tag_field] = batch_df[self.tag_field].tolist()
        return return_dict

    def _batch_tokenize(self, batch_df, fields):
        return_dict = {}
        for field in fields:
            texts = batch_df[field].tolist()
            tokenized = self.tokenizer(texts, padding=True, truncation=True, max_length=self.max_length, return_tensors='pt')
            return_dict[field] = tokenized
        return return_dict

def batch_to_device(batch, target_device: device):
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], Tensor):
            if 'cuda' in target_device:
                batch[key] = batch[key].cuda()
    return batch
